import re
from functools import lru_cache
from typing import List, Optional, Tuple

import requests


@lru_cache(maxsize=35)
def is_pkg_available(pkg_name: str, channel: str = "conda-forge") -> bool:
    """Verify if the package is available on Anaconda for a specific channel.

    :param pkg_name: Package name
    :param channel: Anaconda channel
    :return: Return True if the package is present on the given channel
    """
    response = requests.get(
        url=f"https://anaconda.org/{channel}/{pkg_name}/files", allow_redirects=False
    )
    return response.status_code == 200


def normalize_pkg_name(pkg_name: str) -> str:
    if is_pkg_available(pkg_name):
        return pkg_name
    if is_pkg_available(pkg_name.replace("-", "_")):
        return pkg_name.replace("-", "_")
    elif is_pkg_available(pkg_name.replace("_", "-")):
        return pkg_name.replace("_", "-")
    return pkg_name


def check_pkgs_availability(
    list_pkgs: List[str], channel: Optional[str] = None
) -> List[Tuple[str, bool]]:
    """Check if the list is

    :param list_pkgs: List with packages name
    :return:
    """
    list_pkgs.sort()
    re_search = re.compile(r"^\s*[a-z0-9\.\-\_]+", re.IGNORECASE)

    result_list = []
    all_pkg = set()
    for pkg in list_pkgs:
        if not pkg:
            continue
        search_result = re_search.search(pkg)
        if not search_result:
            continue

        pkg_name = search_result.group()
        if pkg_name in all_pkg:
            continue

        all_pkg.add(pkg_name)
        if channel:
            result_list.append((pkg, is_pkg_available(pkg_name, channel)))
        else:
            result_list.append((pkg, is_pkg_available(pkg_name)))
    return result_list



import pytest

from grayskull.cli.parser import parse_pkg_name_version


def test_parse_pkg_name_version():
    origin, pkg_name, pkg_version = parse_pkg_name_version("pytest=5.3.5")
    assert pkg_name == "pytest"
    assert pkg_version == "5.3.5"

    origin, pkg_name, pkg_version = parse_pkg_name_version("pytest==5.3.5")
    assert pkg_name == "pytest"
    assert pkg_version == "5.3.5"

    origin, pkg_name, pkg_version = parse_pkg_name_version("pytest")
    assert pkg_name == "pytest"
    assert pkg_version is None


def test_parse_git_github_url():
    origin, pkg_name, version = parse_pkg_name_version(
        "https://github.com/pytest-dev/pytest.git"
    )
    assert origin == "https://github.com/pytest-dev/"
    assert pkg_name == "pytest"
    assert not version


@pytest.mark.parametrize("extension", [".zip", ".tar", ".tar.gz", ".tar.bz2"])
@pytest.mark.parametrize(
    "filepath, expected_name, expected_version",
    [
        ("mypkg-1.2.0", "mypkg", "1.2.0"),
        ("mypkg-with-dash-1.2.0", "mypkg-with-dash", "1.2.0"),
        ("mypkg-1.0rc1", "mypkg", "1.0rc1"),
        ("mypkg", "mypkg", ""),
    ],
)
def test_parse_local_sdist(
    extension, filepath, expected_name, expected_version, tmp_path
):
    p = tmp_path / f"{filepath}{extension}"
    p.write_text("foo")
    origin, name, version = parse_pkg_name_version(str(p))
    assert origin == ""
    assert name == expected_name
    assert version == expected_version



import json
import logging
import os
import re
from pathlib import Path
from tempfile import mkdtemp
from typing import Dict, List, Optional

import requests
from colorama import Fore
from souschef.jinja_expression import set_global_jinja_var
from souschef.recipe import Recipe

from grayskull.base.github import generate_git_archive_tarball_url, handle_gh_version
from grayskull.base.pkg_info import normalize_pkg_name
from grayskull.base.track_packages import _get_track_info_from_file, solve_list_pkg_name
from grayskull.cli.stdout import print_msg, print_requirements, progressbar_with_status
from grayskull.config import Configuration
from grayskull.strategy.abstract_strategy import AbstractStrategy
from grayskull.strategy.py_base import (
    RE_DEPS_NAME,
    clean_deps_for_conda_forge,
    discover_license,
    ensure_pep440,
    ensure_pep440_in_req_list,
    get_compilers,
    get_entry_points_from_sdist,
    get_extra_from_requires_dist,
    get_name_version_from_requires_dist,
    get_sdist_metadata,
    get_test_entry_points,
    get_test_imports,
    parse_extra_metadata_to_selector,
    py_version_to_limit_python,
    py_version_to_selector,
    update_requirements_with_pin,
)
from grayskull.utils import format_dependencies, origin_is_github, rm_duplicated_deps

log = logging.getLogger(__name__)

PYPI_CONFIG = Path(os.path.dirname(__file__)) / "config.yaml"

ALL_SECTIONS = (
    "package",
    "source",
    "build",
    "outputs",
    "requirements",
    "app",
    "test",
    "about",
    "extra",
)


class PypiStrategy(AbstractStrategy):
    @staticmethod
    def fetch_data(recipe, config, sections=None):
        update_recipe(recipe, config, sections or ALL_SECTIONS)
        if not (recipe["build"] and recipe["build"]["script"]):
            recipe["build"]["script"] = "<{ PYTHON }} -m pip install . -vv"


def merge_pypi_sdist_metadata(
    pypi_metadata: dict, sdist_metadata: dict, config: Configuration
) -> dict:
    """This method is responsible to merge two dictionaries and it will give
    priority to the pypi_metadata."""

    def get_val(key):
        return pypi_metadata.get(key) or sdist_metadata.get(key)

    requires_dist = merge_requires_dist(pypi_metadata, sdist_metadata)
    all_packages_names = get_val("packages")
    if not all_packages_names:
        all_packages_names = get_val("py_modules")

    source_section = get_val("source")
    if not config.from_local_sdist:
        source_section["url"] = adjust_source_url_to_include_placeholders(
            source_section["url"], get_val("version")
        )

    return {
        "author": get_val("author"),
        "name": get_val("name"),
        "version": get_val("version"),
        "source": source_section,
        "packages": all_packages_names,
        "url": get_val("url"),
        "classifiers": get_val("classifiers"),
        "compilers": get_compilers(requires_dist, sdist_metadata, config),
        "entry_points": get_entry_points_from_sdist(sdist_metadata),
        "scripts": get_val("scripts"),
        "summary": get_val("summary"),
        "requires_python": ensure_pep440(
            pypi_metadata.get("requires_python")
            or sdist_metadata.get("python_requires")
        ),
        "doc_url": get_val("doc_url"),
        "dev_url": get_val("dev_url"),
        "license": get_val("license"),
        "setup_requires": get_val("setup_requires"),
        "extra_requires": get_val("extra_requires"),
        "project_url": get_val("project_url"),
        "extras_require": get_val("extras_require"),
        "requires_dist": requires_dist,
        "sdist_path": get_val("sdist_path"),
    }


def adjust_source_url_to_include_placeholders(url, version):
    url_split = url.rsplit("/", 1)
    url_split[-1] = url_split[-1].replace(version, "{{ version }}")
    return "/".join(url_split)


def get_url_filename(metadata: dict, default: Optional[str] = None) -> str:
    """Method responsible to get the filename and right extension to add
    to the pypi url

    :param metadata: Dictionary with the all package metadata filled
    :param default: default value for the package filename
    :return: filename and extension to download the file on pypi
    """
    if default is None:
        default = "{{ name }}-{{ version }}.tar.gz"
    if "urls" not in metadata:
        return default

    for pkg_url in metadata["urls"]:
        if pkg_url["packagetype"] == "sdist":
            version = metadata["info"]["version"]
            return pkg_url["filename"].replace(version, "{{ version }}")
    return default


def get_sha256_from_pypi_metadata(pypi_metadata: dict) -> str:
    """Get the sha256 from pypi metadata

    :param pypi_metadata: pypi metadata
    :return: sha256 value for the sdist package
    """
    for pkg_info in pypi_metadata.get("urls"):
        if pkg_info.get("packagetype", "") == "sdist":
            return pkg_info["digests"]["sha256"]
    raise AttributeError("Hash information for sdist was not found on PyPi metadata.")


def get_sdist_url_from_pypi(metadata: dict) -> str:
    """Return the sdist url looking for the pypi metadata

    :param metadata: pypi metadata
    :return: sdist url
    """
    for sdist_url in metadata["urls"]:
        if sdist_url["packagetype"] == "sdist":
            return sdist_url["url"]


def skip_pypi_requirement(list_extra: List) -> bool:
    """Test if it should skip the requirement

    :param list_extra: list with all extra requirements
    :return: True if we should skip the requirement
    """
    return any(extra[1] == "extra" or extra[3] == "testing" for extra in list_extra)


def merge_requires_dist(pypi_metadata: dict, sdist_metadata: dict) -> List:
    """Merge requirements metadata from pypi and sdist.

    :param pypi_metadata: pypi metadata
    :param sdist_metadata: sdist metadata
    :return: list with all requirements
    """
    all_deps = sdist_metadata.get("install_requires") or []
    all_deps += pypi_metadata.get("requires_dist") or []

    re_search = re.compile(r";\s*extra")
    all_deps = [pkg for pkg in all_deps if not re_search.search(pkg)]
    current_pkg = pypi_metadata.get("name", "")

    requires_dist = []
    pypi_deps_name = set()
    with progressbar_with_status(len(all_deps)) as bar:
        for pos, sdist_pkg in enumerate(all_deps, 1):
            match_deps = RE_DEPS_NAME.match(sdist_pkg)
            if not match_deps:
                bar.update(pos)
                continue
            match_deps = match_deps.group(0).strip()
            pkg_name = normalize_pkg_name(match_deps)
            bar.update(pos, pkg_name=pkg_name)
            if current_pkg and current_pkg == pkg_name:
                continue
            if pkg_name in pypi_deps_name:
                continue

            pypi_deps_name.add(pkg_name)
            requires_dist.append(sdist_pkg.replace(match_deps, pkg_name))
    return requires_dist


def get_origin_wise_metadata(config):
    """Method responsible for extracting metadata based on package origin."""
    if config.repo_github and origin_is_github(config.repo_github):
        url = config.repo_github
        name = config.name
        version, version_tag = handle_gh_version(
            name=name, version=config.version, url=url, tag=config.github_release_tag
        )
        archive_url = generate_git_archive_tarball_url(git_url=url, git_ref=version_tag)
        sdist_metadata = get_sdist_metadata(
            sdist_url=archive_url, config=config, with_source=True
        )
        sdist_metadata["version"] = version
        pypi_metadata = {}
    elif config.from_local_sdist:
        sdist_metadata = get_sdist_metadata(
            sdist_url="",
            config=config,
        )
        pypi_metadata = {}
    else:
        pypi_metadata = get_pypi_metadata(config)
        sdist_metadata = get_sdist_metadata(
            sdist_url=pypi_metadata["sdist_url"], config=config
        )
    return sdist_metadata, pypi_metadata


def get_pypi_metadata(config: Configuration) -> dict:
    """Method responsible to communicate with the pypi api endpoints and
    get the whole metadata available for the specified package and version.

    :return: Pypi metadata
    """
    print_msg("Recovering metadata from pypi...")
    if config.version:
        url_pypi = config.url_pypi_metadata.format(
            pkg_name=f"{config.name}/{config.version}"
        )
    else:
        log.info(f"Version for {config.name} not specified.\nGetting the latest one.")
        url_pypi = config.url_pypi_metadata.format(pkg_name=config.name)

    metadata = requests.get(url=url_pypi, timeout=5)
    if metadata.status_code != 200:
        raise requests.HTTPError(
            f"It was not possible to recover package metadata for {config.name}.\n"
            f"Error code: {metadata.status_code}"
        )

    metadata = metadata.json()
    if config.download:
        download_file = os.path.join(
            str(mkdtemp(f"grayskull-pypi-metadata-{config.name}-")), "pypi.json"
        )
        with open(download_file, "w") as f:
            json.dump(metadata, f, indent=4)
        config.files_to_copy.append(download_file)
    info = metadata["info"]
    project_urls = info.get("project_urls") or {}
    log.info(f"Package: {config.name}=={info['version']}")
    log.debug(f"Full PyPI metadata:\n{metadata}")
    sdist_url = get_sdist_url_from_pypi(metadata)
    if sdist_url is None:
        raise AttributeError(f"There is no sdist package on pypi for {config.name}.")
    return {
        "name": config.name,
        "version": info["version"],
        "requires_dist": info.get("requires_dist", []),
        "requires_python": info.get("requires_python", None),
        "summary": info.get("summary"),
        "project_url": info.get("project_url"),
        "doc_url": info.get("docs_url"),
        "dev_url": project_urls.get("Source"),
        "url": info.get("home_page"),
        "license": info.get("license"),
        "source": {
            "url": "https://pypi.io/packages/source/{{ name[0] }}/{{ name }}/"
            f"{get_url_filename(metadata)}",
            "sha256": get_sha256_from_pypi_metadata(metadata),
        },
        "sdist_url": sdist_url,
    }


def get_run_req_from_requires_dist(requires_dist: List, config: Configuration) -> List:
    """Get the run requirements looking for the `requires_dist` key
    present in the metadata"""
    run_req = []
    for req in requires_dist:
        list_raw_requirements = req.split(";")
        selector = ""
        if len(list_raw_requirements) > 1:
            list_extra = get_extra_from_requires_dist(list_raw_requirements[1])
            if skip_pypi_requirement(list_extra):
                continue

            if result_selector := get_all_selectors_pypi(list_extra, config):
                selector = " ".join(result_selector)
                selector = f"  # [{selector}]"
            else:
                selector = ""
        pkg_name, version = get_name_version_from_requires_dist(
            list_raw_requirements[0]
        )
        run_req.append(f"{pkg_name} {version}{selector}".strip())
    return run_req


def get_all_selectors_pypi(list_extra: List, config: Configuration) -> List:
    """Get the selectors looking for the pypi data

    :param list_extra: List of extra requirements from pypi
    :param config: Configuration instructions for recipe
    :return: List of extra requirements with the selectors
    """
    result_selector = []
    for extra in list_extra:
        config.is_arch = True
        if selector := parse_extra_metadata_to_selector(extra[1], extra[2], extra[3]):
            if extra[0]:
                result_selector.append(extra[0])
            result_selector.append(selector)
            if extra[4]:
                result_selector.append(extra[4])
            if extra[5]:
                result_selector.append(extra[5])
    if result_selector and result_selector[-1] in ["and", "or"]:
        del result_selector[-1]
    return result_selector


def get_metadata(recipe, config) -> dict:
    """Method responsible to get the whole metadata available. It will
    merge metadata from multiple sources (pypi, setup.py, setup.cfg)
    """
    name = config.name
    sdist_metadata, pypi_metadata = get_origin_wise_metadata(config)
    metadata = merge_pypi_sdist_metadata(pypi_metadata, sdist_metadata, config)
    log.debug(f"Data merged from pypi, setup.cfg and setup.py: {metadata}")
    if metadata.get("scripts") is not None:
        config.is_arch = True
        print_msg(f"{Fore.YELLOW}scripts detected. Package not eligible for noarch.")

    all_license_metadata = discover_license(metadata)
    license_file = []
    all_license_name = set()
    for license_metadata in all_license_metadata:
        all_license_name.add(license_metadata.name)
        if license_metadata.path:
            if license_metadata.is_packaged:
                license_file.append(license_metadata.path)
            else:
                license_file.append(os.path.basename(license_metadata.path))
                config.files_to_copy.append(license_metadata.path)

        print_msg(f"License type: {Fore.LIGHTMAGENTA_EX}{license_metadata.name}")
        print_msg(f"License file: {Fore.LIGHTMAGENTA_EX}{license_file}")
    if all_license_name:
        license_name = " & ".join(all_license_name)
        if len(all_license_name) > 1:
            print_msg(
                f"{Fore.RED}Multiple licenses detected!"
                f" {Fore.LIGHTYELLOW_EX}Please, be sure to check them."
            )
    else:
        license_name = "Other"
    if not license_file:
        license_file = ["PLEASE_ADD_LICENSE_FILE"]
    if not license_name:
        license_name = "Other"

    requirements_section = extract_requirements(metadata, config, recipe)
    optional_requirements = extract_optional_requirements(metadata, config)
    for key in requirements_section:
        requirements_section[key] = normalize_requirements_list(
            requirements_section[key], config
        )
    for key in optional_requirements:
        optional_requirements[key] = normalize_requirements_list(
            optional_requirements[key], config
        )

    all_missing_deps = print_requirements(requirements_section, optional_requirements)
    config.missing_deps = all_missing_deps

    test_requirements = optional_requirements.pop(config.extras_require_test, [])
    test_section = compose_test_section(metadata, test_requirements)

    about_section = {
        "home": metadata["url"] if metadata.get("url") else metadata.get("project_url"),
        "summary": metadata.get("summary"),
        "doc_url": metadata.get("doc_url"),
        "dev_url": metadata.get("dev_url"),
        "license": license_name,
        "license_file": license_file,
    }

    source_section = metadata.get("source", {})

    build_section = {"entry_points": metadata.get("entry_points")}

    outputs = []
    if optional_requirements:
        if config.extras_require_split:
            outputs.append(
                {
                    "name": name,
                    "requirements": requirements_section,
                }
            )
            for option, req_list in optional_requirements.items():
                req_section = dict(requirements_section)
                req_section["run"] = [f"{name} =={{{{ version }}}}"] + req_list
                outputs.append(
                    {
                        "name": f"{name}-{option}",
                        "requirements": req_section,
                    }
                )
        else:
            # Sort options in terms of inclusion
            optional_requirements_items = list()
            for option, req_list in optional_requirements.items():
                for _idx, (_, req_list2) in enumerate(optional_requirements_items):
                    if set(req_list).issubset(req_list2):
                        break
                else:
                    _idx = len(optional_requirements_items)
                optional_requirements_items.insert(_idx, (option, req_list))

            # Add options to run requirements
            for option, req_list in optional_requirements_items:
                req_list = [s for s in req_list if s not in requirements_section["run"]]
                if not req_list:
                    continue
                requirements_section["run"] += [f"# Extra: {option}"] + req_list

    if outputs:
        package_section = {"name": name + "-meta", "version": metadata["version"]}
        return {
            "package": package_section,
            "build": build_section,
            "outputs": outputs,
            "test": test_section,
            "about": about_section,
            "source": source_section,
        }
    else:
        package_section = {"name": name, "version": metadata["version"]}
        return {
            "package": package_section,
            "build": build_section,
            "requirements": requirements_section,
            "test": test_section,
            "about": about_section,
            "source": source_section,
        }


def update_recipe(recipe: Recipe, config: Configuration, all_sections: List[str]):
    """Update one specific section."""
    from souschef.section import Section

    metadata = get_metadata(recipe, config)
    for section in all_sections:
        if metadata.get(section):
            if section == "package":
                set_global_jinja_var(recipe, "version", metadata["package"]["version"])
                config.version = metadata["package"]["version"]
                recipe["package"]["version"] = "<{ version }}"
            elif section in recipe and isinstance(recipe[section], Section):
                recipe[section].update(metadata[section])
            else:
                recipe.add_section({section: metadata[section]})

    if not config.is_arch:
        recipe["build"]["noarch"] = "python"


def extract_requirements(metadata: dict, config, recipe) -> Dict[str, List[str]]:
    """Extract the requirements for `build`, `host` and `run`"""
    name = metadata["name"]
    requires_dist = format_dependencies(metadata.get("requires_dist", []), name)
    setup_requires = metadata.get("setup_requires", [])
    host_req = format_dependencies(setup_requires or [], config.name)
    if not requires_dist and not host_req and not metadata.get("requires_python"):
        if config.is_strict_cf:
            return {
                "host": ["python", "pip"],
                "run": [
                    f"python"
                    f" >={config.py_cf_supported[0].major}"
                    f".{config.py_cf_supported[0].minor}"
                ],
            }
        else:
            return {"host": ["python", "pip"], "run": ["python"]}

    run_req = get_run_req_from_requires_dist(requires_dist, config)
    build_req = [f"<{{ compiler('{c}') }}}}" for c in metadata.get("compilers", [])]
    if build_req:
        config.is_arch = True

    if config.is_arch:
        if version_to_selector := py_version_to_selector(metadata, config):
            try:
                recipe["build"]["skip"] = True
            except (TypeError, KeyError):
                recipe.add_section({"build": {"skip": True}})
            recipe["build"]["skip"].inline_comment = version_to_selector
        limit_python = ""
    else:
        limit_python = py_version_to_limit_python(metadata, config)

    limit_python = f" {limit_python}" if limit_python else ""

    if "pip" not in host_req:
        host_req += [f"python{limit_python}", "pip"]
    run_req.insert(0, f"python{limit_python}")

    if config.is_strict_cf:
        host_req = remove_selectors_pkgs_if_needed(host_req)
        run_req = remove_selectors_pkgs_if_needed(run_req)
    result = {}
    if build_req:
        result = {
            "build": rm_duplicated_deps(sorted(map(lambda x: x.lower(), build_req)))
        }

    result.update(
        {
            "host": rm_duplicated_deps(sorted(map(lambda x: x.lower(), host_req))),
            "run": rm_duplicated_deps(sorted(map(lambda x: x.lower(), run_req))),
        }
    )
    update_requirements_with_pin(result)
    return result


def remove_selectors_pkgs_if_needed(
    list_req: List, config_file: Optional[Path] = None
) -> List:
    info_pkgs = _get_track_info_from_file(config_file or PYPI_CONFIG)
    re_selector = re.compile(r"\s+#\s+\[.*", re.DOTALL)
    result = []
    for pkg in list_req:
        pkg_cfg_info = info_pkgs.get(pkg.strip().split()[0], {})
        if pkg_cfg_info.get("avoid_selector", False):
            pkg = re_selector.sub("", pkg)
        result.append(pkg)
    return result


def extract_optional_requirements(metadata: dict, config) -> Dict[str, List[str]]:
    """Extract all optional requirements that are specified in the configuration"""
    keys = set()
    if config.extras_require_all:
        extras_require = metadata.get("extras_require")
        if extras_require:
            keys.update(extras_require.keys())
    if config.extras_require_include:
        keys.update(config.extras_require_include)
    if config.extras_require_exclude:
        keys -= set(config.extras_require_exclude)
    if config.extras_require_test:
        keys.add(config.extras_require_test)
    result = dict()
    for key in sorted(keys):
        requirements = metadata["extras_require"].get(key)
        if requirements:
            result[key] = requirements
    return result


def normalize_requirements_list(requirements: List[str], config) -> List[str]:
    """Adapt requirements to PEP440, Conda and Conda-Forge"""
    requirements = solve_list_pkg_name(requirements, PYPI_CONFIG)
    requirements = ensure_pep440_in_req_list(requirements)
    if config.is_strict_cf:
        requirements = clean_deps_for_conda_forge(
            requirements, config.py_cf_supported[0]
        )
    return requirements


def compose_test_section(metadata: dict, test_requirements: List[str]) -> dict:
    test_imports = get_test_imports(metadata, metadata["name"])
    test_requirements = ["pip"] + test_requirements
    test_commands = ["pip check"]
    if any("pytest" in req for req in test_requirements):
        test_commands.extend(f"pytest --pyargs {module}" for module in test_imports)
    test_commands.extend(get_test_entry_points(metadata.get("entry_points", [])))
    return {
        "imports": test_imports,
        "commands": test_commands,
        "requires": test_requirements,
    }



import sys
from copy import deepcopy

import pytest
from souschef.recipe import Recipe

import grayskull
from grayskull import __main__ as cli
from grayskull.base.factory import GrayskullFactory
from grayskull.config import Configuration


def test_version(capsys):
    cli.main(["--version"])
    captured = capsys.readouterr()
    assert captured.out.strip() == grayskull.__version__


def test_pypi_cmd(tmpdir):
    out_folder = tmpdir.mkdir("out")
    cli.main(
        ["pypi", "pytest=5.3.2", "-o", str(out_folder), "-m", "m1", "m2", "--download"]
    )
    pytest_folder = out_folder / "pytest"
    assert pytest_folder.isdir()

    recipe_file = pytest_folder / "meta.yaml"
    assert recipe_file.isfile()
    assert (pytest_folder / "pypi.json").isfile()
    assert (pytest_folder / "pytest-5.3.2.tar.gz").isfile()


@pytest.mark.parametrize("option", ["--heman", "--shera"])
def test_easter(capsys, option):
    cli.main([option])
    captured = capsys.readouterr()
    assert "By the power of Grayskull..." in captured.out.strip()


def test_grayskull_without_options(capsys, monkeypatch):
    monkeypatch.setattr(sys, "argv", ["foo"])
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        cli.main([])
    assert pytest_wrapped_e.type == SystemExit
    captured = capsys.readouterr()
    assert "Grayskull - Conda recipe generator" in captured.out


def test_msg_missing_pkg_pypi(capsys):
    cli.main(["pypi", "NOT_A_PACKAGE_123123123"])
    captured = capsys.readouterr()
    assert (
        "Package seems to be missing."
        "\nException: It was not possible to recover package metadata"
        " for NOT_A_PACKAGE_123123123.\n"
        "Error code: 404" in captured.out
    )


def test_license_discovery(tmpdir):
    out_folder = tmpdir.mkdir("out-license")
    cli.main(["pypi", "httplib2shim=0.0.3", "-o", str(out_folder)])
    assert (out_folder / "httplib2shim" / "LICENSE").exists()


def test_change_pypi_url(mocker):
    mocker.patch("grayskull.__main__.generate_recipe", return_value=None)
    mocker.patch(
        "grayskull.__main__.create_python_recipe", return_value=({"extra": {}}, None)
    )
    mocker.patch("grayskull.__main__.add_extra_section", return_value=None)
    spy = mocker.spy(cli, "create_python_recipe")

    cli.main(["pypi", "pytest=5.3.2", "--pypi-url=http://url_pypi.com/abc"])
    spy.assert_called_once_with(
        "pytest=5.3.2",
        is_strict_cf=False,
        download=False,
        url_pypi_metadata="http://url_pypi.com/abc",
        sections_populate=None,
        from_local_sdist=False,
        extras_require_test=None,
        github_release_tag=None,
        extras_require_include=tuple(),
        extras_require_exclude=tuple(),
        extras_require_all=False,
        extras_require_split=False,
    )


def test_config_url_pypi_metadata():
    config = Configuration("pytest", url_pypi_metadata="http://url_pypi.com/abc")
    assert config.url_pypi_metadata == "http://url_pypi.com/abc/{pkg_name}/json"


@pytest.mark.parametrize("option", ["-r", "--recursive"])
def test_recursive_option(mocker, option, tmpdir):
    folder = tmpdir.mkdir(f"recursive_pkg{option}")

    def mock_is_pkg_available(pkg):
        return pkg != "colorama"

    mocker.patch("grayskull.cli.stdout.is_pkg_available", new=mock_is_pkg_available)
    spy = mocker.spy(cli, "generate_recipes_from_list")
    cli.main(["pypi", "pytest=5.3.2", option, "-o", str(folder)])
    assert spy.call_count == 2
    assert spy.call_args_list[0].args[0] == ["pytest=5.3.2"]
    assert spy.call_args_list[1].args[0] == {"colorama"}


def test_part_reload_recipe(tmpdir):
    recipe = GrayskullFactory.create_recipe(
        "pypi", Configuration(name="pytest", version="5.3.2")
    )
    host = deepcopy([str(i) for i in recipe["requirements"]["host"]])
    run = deepcopy([str(i) for i in recipe["requirements"]["run"]])
    recipe["requirements"] = {}
    recipe["foo"] = "bar"
    assert not recipe["requirements"].value
    assert host
    assert run
    assert recipe["foo"] == "bar"

    folder = tmpdir.mkdir("reload_recipe")
    recipe_path = folder / "recipe.yaml"
    recipe.save(str(recipe_path))
    cli.main(["pypi", str(recipe_path), "--sections", "requirements"])

    recipe = Recipe(load_file=str(recipe_path))
    assert host == [str(v) for v in recipe["requirements"]["host"] if v]
    assert run == [str(v) for v in recipe["requirements"]["run"] if v]
    assert recipe["foo"] == "bar"



import os
import shutil

import pytest
from pytest import fixture

from grayskull.strategy.py_base import download_sdist_pkg


@fixture(scope="session")
def data_dir() -> str:
    return os.path.join(os.path.dirname(__file__), "data")


@fixture(scope="session")
def pkg_pytest(tmpdir_factory) -> str:
    folder = tmpdir_factory.mktemp("test-download-pkg")
    dest_pkg = str(folder / "PYTEST-PKG.tar.gz")
    download_sdist_pkg(
        "https://pypi.io/packages/source/p/pytest/pytest-5.3.5.tar.gz", dest_pkg
    )
    shutil.unpack_archive(dest_pkg, str(folder))
    return dest_pkg


def pytest_collection_modifyitems(config, items):
    github_mark = pytest.mark.xfail(
        reason="This test may fail because github has limitation regarding the"
        " number of requisitions we can do to their api."
    )
    for item in items:
        if "github" in item.keywords:
            item.add_marker(github_mark)



from grayskull.base.pkg_info import check_pkgs_availability, is_pkg_available


def test_pkg_available():
    assert is_pkg_available("pytest")


def test_pkg_not_available():
    assert not is_pkg_available("NOT_PACKAGE_654987321")


def test_check_pkgs_availability(capsys):
    all_pkgs = check_pkgs_availability(
        [
            "pytest >=4.0.0,<5.0.0  # [win]",
            "requests",
            "pandas[test]",
            "NOT_PACKAGE_13248743",
        ]
    )

    assert sorted(all_pkgs) == sorted(
        [
            ("pytest >=4.0.0,<5.0.0  # [win]", True),
            ("requests", True),
            ("pandas[test]", True),
            ("NOT_PACKAGE_13248743", False),
        ]
    )



import os

import pytest

from grayskull.utils import (
    format_dependencies,
    get_all_modules_imported_script,
    get_std_modules,
    get_vendored_dependencies,
    merge_dict_of_lists_item,
    merge_list_item,
    origin_is_local_sdist,
)


def test_get_std_modules():
    std_modules = get_std_modules()
    assert "sys" in std_modules
    assert "os" in std_modules
    assert "ast" in std_modules
    assert "typing" in std_modules


def test_get_all_modules_imported_script(data_dir):
    all_imports = get_all_modules_imported_script(
        os.path.join(data_dir, "foo_imports.py")
    )
    assert sorted(all_imports) == sorted(["numpy", "pandas", "requests", "os", "sys"])


def test_get_vendored_dependencies(data_dir):
    all_deps = get_vendored_dependencies(os.path.join(data_dir, "foo_imports.py"))
    assert sorted(all_deps) == sorted(["numpy", "pandas", "requests"])


def test_format_dependencies_optional_double_equal():
    assert format_dependencies(
        ["dask[dataframe,distributed]==2021.10.0"], "dask-sql"
    ) == ["dask ==2021.10.0"]


@pytest.mark.parametrize(
    "filename", ["mypkg.zip", "mypkg.tar", "mypkg.tar.gz", "mypkg.tar.bz2"]
)
def test_origin_is_local_sdist(filename, tmp_path):
    p = tmp_path / filename
    p.write_text("foo")
    assert origin_is_local_sdist(str(p))


def test_origin_is_not_local_sdist_file(tmp_path):
    p = tmp_path / "mypgk"
    p.write_text("foo")
    assert not origin_is_local_sdist(str(p))


@pytest.mark.parametrize(
    "filename", ["mypkg", "mypkg=1.0.0", "https://github.com/foo/bar"]
)
def test_origin_is_not_local_sdist_filename(filename):
    assert not origin_is_local_sdist(filename)


def test_merge_lists_item():
    destination = {}
    add = {}
    merge_list_item(destination, add, "name")
    assert destination == {}

    destination = {"name": [1]}
    add = {}
    merge_list_item(destination, add, "name")
    destination = {key: set(lst) for key, lst in destination.items()}
    assert destination == {"name": {1}}

    destination = {}
    add = {"name": [2]}
    merge_list_item(destination, add, "name")
    destination = {key: set(lst) for key, lst in destination.items()}
    assert destination == {"name": {2}}

    destination = {"name": [1]}
    add = {"name": [2]}
    merge_list_item(destination, add, "name")
    destination = {key: set(lst) for key, lst in destination.items()}
    assert destination == {"name": {1, 2}}


def test_merge_dict_of_lists_item():
    destination = {}
    add = {}
    merge_dict_of_lists_item(destination, add, "name")
    assert destination == {}

    destination = {"name": {"sub_name": [1]}}
    add = {}
    merge_dict_of_lists_item(destination, add, "name")
    for key in destination:
        destination[key] = {
            sub_key: set(lst) for sub_key, lst in destination[key].items()
        }
    assert destination == {"name": {"sub_name": {1}}}

    destination = {}
    add = {"name": {"sub_name": [2]}}
    merge_dict_of_lists_item(destination, add, "name")
    for key in destination:
        destination[key] = {
            sub_key: set(lst) for sub_key, lst in destination[key].items()
        }
    assert destination == {"name": {"sub_name": {2}}}

    destination = {"name": {"sub_name": [1]}}
    add = {"name": {"sub_name": [2]}}
    merge_dict_of_lists_item(destination, add, "name")
    for key in destination:
        destination[key] = {
            sub_key: set(lst) for sub_key, lst in destination[key].items()
        }
    assert destination == {"name": {"sub_name": {1, 2}}}



import hashlib
import json
import os
import sys
from unittest.mock import patch

import pytest
from colorama import Fore, Style
from souschef.recipe import Recipe

from grayskull.__main__ import create_python_recipe
from grayskull.base.factory import GrayskullFactory
from grayskull.base.pkg_info import normalize_pkg_name
from grayskull.cli import CLIConfig
from grayskull.cli.parser import parse_pkg_name_version
from grayskull.config import Configuration
from grayskull.strategy.py_base import (
    clean_deps_for_conda_forge,
    ensure_pep440,
    generic_py_ver_to,
    get_compilers,
    get_entry_points_from_sdist,
    get_extra_from_requires_dist,
    get_name_version_from_requires_dist,
    get_sdist_metadata,
    get_setup_cfg,
    get_test_entry_points,
    get_test_imports,
    parse_extra_metadata_to_selector,
    py_version_to_limit_python,
    py_version_to_selector,
    update_requirements_with_pin,
)
from grayskull.strategy.pypi import (
    PypiStrategy,
    compose_test_section,
    extract_optional_requirements,
    extract_requirements,
    get_all_selectors_pypi,
    get_pypi_metadata,
    get_sha256_from_pypi_metadata,
    get_url_filename,
    merge_pypi_sdist_metadata,
    normalize_requirements_list,
    remove_selectors_pkgs_if_needed,
)
from grayskull.utils import PyVer, format_dependencies, generate_recipe


@pytest.fixture
def pypi_metadata():
    path_metadata = os.path.join(
        os.path.dirname(__file__), "data", "pypi_pytest_metadata.json"
    )
    with open(path_metadata) as f:
        return json.load(f)


@pytest.fixture
def recipe_config():
    config = Configuration(name="pytest")
    recipe = Recipe(name="pytest")
    return recipe, config


@pytest.fixture
def pypi_metadata_with_extras():
    path_metadata = os.path.join(
        os.path.dirname(__file__), "data", "pypi_dask_metadata.json"
    )
    with open(path_metadata) as f:
        return json.load(f)


def test_extract_pypi_requirements(pypi_metadata, recipe_config):
    recipe, config = recipe_config
    pypi_reqs = extract_requirements(pypi_metadata["info"], config, recipe)
    assert sorted(pypi_reqs["host"]) == sorted(["python", "pip"])
    assert sorted(pypi_reqs["run"]) == sorted(
        [
            "python",
            "py >=1.5.0",
            "packaging",
            "attrs >=17.4.0",
            "more-itertools >=4.0.0",
            "pluggy <1.0,>=0.12",
            "wcwidth",
            "pathlib2 >=2.2.0  # [py<36]",
            "importlib-metadata >=0.12  # [py<38]",
            "atomicwrites >=1.0  # [win]",
            "colorama   # [win]",
        ]
    )


def test_get_pypi_metadata(pypi_metadata):
    recipe = Recipe(name="pytest")
    config = Configuration(name="pytest", version="5.3.1", is_strict_cf=True)
    metadata = get_pypi_metadata(config)
    PypiStrategy.fetch_data(recipe, config)
    assert metadata["name"] == "pytest"
    assert metadata["version"] == "5.3.1"
    assert "pathlib2 >=2.2.0  # [py<36]" not in recipe["requirements"]["run"]


def test_get_name_version_from_requires_dist():
    assert get_name_version_from_requires_dist("py (>=1.5.0)") == (
        "py",
        ">=1.5.0",
    )


def test_get_extra_from_requires_dist():
    assert get_extra_from_requires_dist(' python_version < "3.6"') == [
        (
            "",
            "python_version",
            "<",
            "3.6",
            "",
            "",
        )
    ]
    assert get_extra_from_requires_dist(
        " python_version < \"3.6\" ; extra =='test'"
    ) == [
        ("", "python_version", "<", "3.6", "", ""),
        ("", "extra", "==", "test", "", ""),
    ]
    assert get_extra_from_requires_dist(
        ' (sys_platform =="win32" and python_version =="2.7") and extra =="socks"'
    ) == [
        ("(", "sys_platform", "==", "win32", "", "and"),
        ("", "python_version", "==", "2.7", ")", "and"),
        ("", "extra", "==", "socks", "", ""),
    ]


@pytest.fixture(scope="module")
def dask_sdist_metadata():
    config = Configuration(name="dask")
    metadata = get_sdist_metadata(
        "https://pypi.io/packages/source/d/dask/dask-2022.6.1.tar.gz",
        config,
    )
    return metadata


def test_get_extra_requirements(dask_sdist_metadata):
    received = {
        extra: set(req_lst)
        for extra, req_lst in dask_sdist_metadata["extras_require"].items()
    }
    expected = {
        "array": {"numpy >= 1.18"},
        "bag": set(),
        "dataframe": {"pandas >= 1.0", "numpy >= 1.18"},
        "distributed": {"distributed == 2022.6.1"},
        "diagnostics": {"bokeh >= 2.4.2", "jinja2"},
        "delayed": set(),
        "complete": {
            "bokeh >= 2.4.2",
            "numpy >= 1.18",
            "distributed == 2022.6.1",
            "pandas >= 1.0",
            "jinja2",
        },
        "test": {"pytest-xdist", "pytest-rerunfailures", "pre-commit", "pytest"},
    }
    assert received == expected


def test_extract_optional_requirements(dask_sdist_metadata):
    config = Configuration(name="dask")

    received = extract_optional_requirements(dask_sdist_metadata, config)
    assert not received

    all_optional_reqs = {
        "array": {"numpy >= 1.18"},
        "complete": {
            "distributed == 2022.6.1",
            "pandas >= 1.0",
            "numpy >= 1.18",
            "bokeh >= 2.4.2",
            "jinja2",
        },
        "dataframe": {"numpy >= 1.18", "pandas >= 1.0"},
        "diagnostics": {"bokeh >= 2.4.2", "jinja2"},
        "distributed": {"distributed == 2022.6.1"},
        "test": {"pytest-xdist", "pre-commit", "pytest-rerunfailures", "pytest"},
    }

    config.extras_require_all = True
    received = extract_optional_requirements(dask_sdist_metadata, config)
    received = {k: set(v) for k, v in received.items()}
    expected = {k: set(v) for k, v in all_optional_reqs.items()}
    assert received == expected

    config.extras_require_all = True
    config.extras_require_include = None
    config.extras_require_exclude = ["complete"]
    received = extract_optional_requirements(dask_sdist_metadata, config)
    received = {k: set(v) for k, v in received.items()}
    expected = {k: set(v) for k, v in all_optional_reqs.items() if k != "complete"}
    assert received == expected

    config.extras_require_all = False
    config.extras_require_include = ["complete"]
    config.extras_require_exclude = None
    received = extract_optional_requirements(dask_sdist_metadata, config)
    received = {k: set(v) for k, v in received.items()}
    expected = {k: set(v) for k, v in all_optional_reqs.items() if k == "complete"}
    assert received == expected

    config.extras_require_all = True
    config.extras_require_include = ["complete"]
    config.extras_require_exclude = None
    received = extract_optional_requirements(dask_sdist_metadata, config)
    received = {k: set(v) for k, v in received.items()}
    expected = {k: set(v) for k, v in all_optional_reqs.items()}
    assert received == expected

    config.extras_require_all = True
    config.extras_require_include = None
    config.extras_require_exclude = ["complete", "test"]
    received = extract_optional_requirements(dask_sdist_metadata, config)
    received = {k: set(v) for k, v in received.items()}
    expected = {
        k: set(v) for k, v in all_optional_reqs.items() if k not in ("complete", "test")
    }
    assert received == expected

    config.extras_require_all = True
    config.extras_require_include = None
    config.extras_require_exclude = ["complete", "test"]
    config.extras_require_test = "test"
    received = extract_optional_requirements(dask_sdist_metadata, config)
    received = {k: set(v) for k, v in received.items()}
    expected = {k: set(v) for k, v in all_optional_reqs.items() if k != "complete"}
    assert received == expected


def test_compose_test_section_with_console_scripts():
    config = Configuration(name="pytest", version="7.1.2")
    metadata1 = get_pypi_metadata(config)
    metadata2 = get_sdist_metadata(
        "https://pypi.io/packages/source/p/pytest/pytest-7.1.2.tar.gz", config
    )
    metadata = merge_pypi_sdist_metadata(metadata1, metadata2, config)
    test_requirements = []
    test_section = compose_test_section(metadata, test_requirements)
    test_section = {k: set(v) for k, v in test_section.items()}
    expected = {
        "imports": {"pytest"},
        "commands": {"pip check", "py.test --help", "pytest --help"},
        "requires": {"pip"},
    }
    assert test_section == expected


def test_compose_test_section_with_requirements(dask_sdist_metadata):
    config = Configuration(name="dask", version="2022.7.1")
    metadata = get_pypi_metadata(config)
    test_requirements = dask_sdist_metadata["extras_require"]["test"]
    test_section = compose_test_section(metadata, test_requirements)
    test_section = {k: set(v) for k, v in test_section.items()}
    expected = {
        "imports": {"dask"},
        "commands": {"pip check", "pytest --pyargs dask"},
        "requires": {
            "pip",
            "pytest",
            "pytest-xdist",
            "pytest-rerunfailures",
            "pre-commit",
        },
    }
    assert test_section == expected


def test_get_include_extra_requirements():
    base_requirements = [
        "cloudpickle >=1.1.1",
        "fsspec >=0.6.0",
        "packaging >=20.0",
        "partd >=0.3.10",
        "python >=3.8",
        "pyyaml >=5.3.1",
        "toolz >=0.8.2",
    ]

    extras = dict()
    extras["array"] = ["numpy >=1.18"]
    extras["distributed"] = ["distributed ==2022.6.1"]
    extras["diagnostics"] = ["bokeh >=2.4.2", "jinja2"]
    extras["dataframe"] = ["numpy >=1.18", "pandas >=1.0"]
    extras["test"] = ["pytest-xdist", "pytest", "pytest-rerunfailures", "pre-commit"]
    extras["complete"] = [
        "distributed ==2022.6.1",
        "jinja2",
        "numpy >=1.18",
        "bokeh >=2.4.2",
        "pandas >=1.0",
    ]

    def set_of_strings(sequence):
        return set(map(str, sequence))

    # extras are not used
    config = Configuration(name="dask", version="2022.6.1")
    recipe = GrayskullFactory.create_recipe("pypi", config)
    recipe["name"] == "dask"
    assert "outputs" not in recipe
    assert set(recipe["requirements"]["run"]) == set(base_requirements)
    assert set(recipe["test"]["requires"]) == {"pip"}

    # all extras are included in the requirements
    config = Configuration(name="dask", version="2022.6.1", extras_require_all=True)
    recipe = GrayskullFactory.create_recipe("pypi", config)
    recipe["name"] == "dask"
    assert "outputs" not in recipe

    expected = list(base_requirements)
    for name, req_lst in extras.items():
        if name != "complete":
            expected.append(f"Extra: {name}")
            expected.extend(req_lst)
    assert set_of_strings(recipe["requirements"]["run"]) == set(expected)
    assert set_of_strings(recipe["test"]["requires"]) == {"pip"}

    # all extras are included in the requirements except for the
    # test requirements which are in the test section
    config = Configuration(
        name="dask",
        version="2022.6.1",
        extras_require_all=True,
        extras_require_test="test",
    )
    recipe = GrayskullFactory.create_recipe("pypi", config)
    recipe["name"] == "dask"
    assert "outputs" not in recipe

    expected = list(base_requirements)
    for name, req_lst in extras.items():
        if name not in ("test", "complete"):
            expected.append(f"Extra: {name}")
            expected.extend(req_lst)
    assert set_of_strings(recipe["requirements"]["run"]) == set(expected)
    assert set_of_strings(recipe["test"]["requires"]) == {"pip", *extras["test"]}

    # only "array" is included in the requirements
    config = Configuration(
        name="dask", version="2022.6.1", extras_require_include=("array",)
    )
    recipe = GrayskullFactory.create_recipe("pypi", config)
    recipe["name"] == "dask"
    assert "outputs" not in recipe
    assert set_of_strings(recipe["requirements"]["run"]) == {
        *base_requirements,
        "Extra: array",
        *extras["array"],
    }
    assert set_of_strings(recipe["test"]["requires"]) == {"pip"}

    # only "test" is included but in the test section
    config = Configuration(
        name="dask",
        version="2022.6.1",
        extras_require_all=True,
        extras_require_exclude=set(extras) - {"test"},
        extras_require_test="test",
    )
    recipe = GrayskullFactory.create_recipe("pypi", config)
    recipe["name"] == "dask"
    assert "outputs" not in recipe
    assert set_of_strings(recipe["requirements"]["run"]) == set(base_requirements)
    assert set_of_strings(recipe["test"]["requires"]) == {"pip", *extras["test"]}

    # only "test" is included in the test section
    config = Configuration(
        name="dask",
        version="2022.6.1",
        extras_require_all=True,
        extras_require_exclude=set(extras) - {"test"},
        extras_require_test="test",
        extras_require_split=True,
    )
    recipe = GrayskullFactory.create_recipe("pypi", config)
    recipe["name"] == "dask"
    assert "outputs" not in recipe
    assert set_of_strings(recipe["requirements"]["run"]) == set(base_requirements)
    assert set_of_strings(recipe["test"]["requires"]) == {"pip", *extras["test"]}

    # all extras have their own output except for the
    # test requirements which are in the test section
    config = Configuration(
        name="dask",
        version="2022.6.1",
        extras_require_all=True,
        extras_require_test="test",
        extras_require_split=True,
    )
    recipe = GrayskullFactory.create_recipe("pypi", config)
    recipe["name"] == "dask-meta"
    assert not list(recipe["requirements"].items())
    assert len(recipe["outputs"]) == 6

    expected = dict()
    expected["dask"] = set(base_requirements)
    for name, req_lst in extras.items():
        if name != "test":
            expected[f"dask-{name}"] = {"dask =={{ version }}", *req_lst}
    found = dict()
    for output in recipe["outputs"]:
        found[output["name"]] = set_of_strings(output["requirements"]["run"])
    assert found == expected

    assert set_of_strings(recipe["test"]["requires"]) == {"pip", *extras["test"]}


def test_normalize_requirements_list():
    config = Configuration(name="pytest")
    requirements = ["pytest ~=5.3.2", "pyqt5"]
    requirements = set(normalize_requirements_list(requirements, config))
    expected = {"pytest >=5.3.2,<5.4.dev0", "pyqt"}
    assert requirements == expected


def test_get_all_selectors_pypi(recipe_config):
    _, config = recipe_config
    config.version = "5.3.1"
    assert get_all_selectors_pypi(
        [
            ("(", "sys_platform", "==", "win32", "", "and"),
            ("", "python_version", "==", "2.7", ")", "and"),
            ("", "extra", "==", "socks", "", ""),
        ],
        config,
    ) == ["(", "win", "and", "py==27", ")"]


def test_get_selector():
    assert parse_extra_metadata_to_selector("extra", "==", "win32") == ""
    assert parse_extra_metadata_to_selector("sys_platform", "==", "win32") == "win"
    assert parse_extra_metadata_to_selector("python_version", "<", "3.6") == "py<36"


@pytest.mark.parametrize(
    "requires_python, exp_selector, ex_cf",
    [
        (">=3.5", "2k", "<36"),
        (">=3.6", "2k", None),
        (">=3.7", "<37", "<37"),
        ("<=3.7", ">=38", ">=38"),
        ("<=3.7.1", ">=38", ">=38"),
        ("<3.7", ">=37", ">=37"),
        (">2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*", "<36", "<36"),
        (">=2.7, !=3.6.*", "==36", "<37"),
        (">3.7", "<38", "<38"),
        (">2.7", "2k", "<36"),
        ("<3", "3k", "skip"),
        ("!=3.7", "==37", "==37"),
        ("~=3.7", "<37", "<37"),
    ],
)
def test_py_version_to_selector(requires_python, exp_selector, ex_cf, recipe_config):
    config = recipe_config[1]
    metadata = {"requires_python": requires_python}
    assert py_version_to_selector(metadata, config) == f"# [py{exp_selector}]"

    if ex_cf != "skip":
        expected = f"# [py{ex_cf}]" if ex_cf else None
        config.is_strict_cf = True
        result = py_version_to_selector(metadata, config)
        if isinstance(expected, str):
            assert expected == result
        else:
            assert expected is result


@pytest.mark.parametrize(
    "requires_python, exp_limit, ex_cf",
    [
        (">=3.5", ">=3.5", ">=3.6"),
        (">=3.6", ">=3.6", ">=3.6"),
        (">=3.7", ">=3.7", ">=3.7"),
        ("<=3.7", "<3.8", "<3.8"),
        ("<=3.7.1", "<3.8", "<3.8"),
        ("<3.7", "<3.7", "<3.7"),
        (">2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*", ">=3.6", ">=3.6"),
        (">=2.7, !=3.6.*", "!=3.6", ">=3.7"),
        (">3.7", ">=3.8", ">=3.8"),
        (">2.7", ">=3.6", ">=3.6"),
        ("<3", "<3.0", "skip"),
        ("!=3.7", "!=3.7", "!=3.7"),
        ("~=3.7", ">=3.7", ">=3.7"),
    ],
)
def test_py_version_to_limit_python(requires_python, exp_limit, ex_cf, recipe_config):
    metadata = {"requires_python": requires_python}
    assert py_version_to_limit_python(metadata, recipe_config[1]) == f"{exp_limit}"

    if ex_cf != "skip":
        config = recipe_config[1]
        config.is_strict_cf = True
        result = py_version_to_limit_python(metadata, config)
        if isinstance(ex_cf, str):
            assert ex_cf == result
        else:
            assert ex_cf is result


def test_get_sha256_from_pypi_metadata():
    metadata = {
        "urls": [
            {"packagetype": "egg", "digests": {"sha256": "23123"}},
            {"packagetype": "sdist", "digests": {"sha256": "1234sha256"}},
        ]
    }
    assert get_sha256_from_pypi_metadata(metadata) == "1234sha256"

    metadata = {
        "urls": [
            {"packagetype": "egg", "digests": {"sha256": "23123"}},
            {"packagetype": "wheel", "digests": {"sha256": "1234sha256"}},
        ]
    }
    with pytest.raises(AttributeError) as err:
        get_sha256_from_pypi_metadata(metadata)
    assert err.match("Hash information for sdist was not found on PyPi metadata.")


@pytest.mark.github
@pytest.mark.parametrize(
    "name", ["hypothesis", "https://github.com/HypothesisWorks/hypothesis"]
)
def test_injection_distutils(name):
    config = Configuration(name="hypothesis")
    data = get_sdist_metadata(
        "https://pypi.io/packages/source/h/hypothesis/hypothesis-5.5.1.tar.gz",
        config,
    )
    assert sorted(data["install_requires"]) == sorted(
        ["attrs>=19.2.0", "sortedcontainers>=2.1.0,<3.0.0"]
    )
    assert data["entry_points"] == {
        "pytest11": ["hypothesispytest = hypothesis.extra.pytestplugin"]
    }
    assert data["version"] == "5.5.1"
    assert data["name"] == "hypothesis"
    assert not data.get("compilers")


def test_injection_distutils_pytest():
    config = Configuration(name="pytest", version="5.3.2")
    data = get_sdist_metadata(
        "https://pypi.io/packages/source/p/pytest/pytest-5.3.2.tar.gz", config
    )
    assert sorted(data["install_requires"]) == sorted(
        [
            "py>=1.5.0",
            "packaging",
            "attrs>=17.4.0",
            "more-itertools>=4.0.0",
            'atomicwrites>=1.0;sys_platform=="win32"',
            'pathlib2>=2.2.0;python_version<"3.6"',
            'colorama;sys_platform=="win32"',
            "pluggy>=0.12,<1.0",
            'importlib-metadata>=0.12;python_version<"3.8"',
            "wcwidth",
        ]
    )
    assert sorted(data["setup_requires"]) == sorted(
        ["setuptools>=40.0", "setuptools_scm"]
    )
    assert not data.get("compilers")


def test_injection_distutils_compiler_gsw():
    config = Configuration(name="gsw", version="3.3.1")
    data = get_sdist_metadata(
        "https://pypi.io/packages/source/g/gsw/gsw-3.3.1.tar.gz", config
    )
    assert data.get("compilers") == ["c"]
    assert data["packages"] == ["gsw"]


def test_injection_distutils_setup_reqs_ensure_list():
    pkg_name, pkg_ver = "pyinstaller-hooks-contrib", "2020.7"
    config = Configuration(name=pkg_name, version=pkg_ver)
    data = get_sdist_metadata(
        f"https://pypi.io/packages/source/p/{pkg_name}/{pkg_name}-{pkg_ver}.tar.gz",
        config,
    )
    assert data.get("setup_requires") == ["setuptools >= 30.3.0"]


def test_merge_pypi_sdist_metadata():
    config = Configuration(name="gsw", version="3.3.1")
    pypi_metadata = get_pypi_metadata(config)
    sdist_metadata = get_sdist_metadata(pypi_metadata["sdist_url"], config)
    merged_data = merge_pypi_sdist_metadata(pypi_metadata, sdist_metadata, config)
    assert merged_data["compilers"] == ["c"]
    assert sorted(merged_data["setup_requires"]) == sorted(["numpy"])


def test_update_requirements_with_pin():
    req = {
        "build": ["<{ compiler('c') }}"],
        "host": ["python", "numpy"],
        "run": ["python", "numpy"],
    }
    update_requirements_with_pin(req)
    assert req == {
        "build": ["<{ compiler('c') }}"],
        "host": ["python", "numpy"],
        "run": ["python", "<{ pin_compatible('numpy') }}"],
    }


def test_get_compilers():
    config = Configuration(name="any_package")
    assert get_compilers(["pybind11"], {}, config) == ["cxx"]
    assert get_compilers(["cython"], {}, config) == ["c"]
    assert sorted(get_compilers(["pybind11", "cython"], {}, config)) == sorted(
        ["cxx", "c"]
    )
    assert sorted(get_compilers(["pybind11"], {"compilers": ["c"]}, config)) == sorted(
        ["cxx", "c"]
    )


def test_get_entry_points_from_sdist():
    assert get_entry_points_from_sdist({}) == []
    assert get_entry_points_from_sdist(
        {"entry_points": {"console_scripts": ["console_scripts=entrypoints"]}}
    ) == ["console_scripts=entrypoints"]
    assert get_entry_points_from_sdist(
        {"entry_points": {"gui_scripts": ["gui_scripts=entrypoints"]}}
    ) == ["gui_scripts=entrypoints"]

    assert sorted(
        get_entry_points_from_sdist(
            {
                "entry_points": {
                    "gui_scripts": ["gui_scripts=entrypoints"],
                    "console_scripts": ["console_scripts=entrypoints"],
                }
            }
        )
    ) == sorted(["gui_scripts=entrypoints", "console_scripts=entrypoints"])
    assert sorted(
        get_entry_points_from_sdist(
            {
                "entry_points": {
                    "gui_scripts": None,
                    "console_scripts": "console_scripts=entrypoints",
                }
            }
        )
    ) == sorted(["console_scripts=entrypoints"])
    assert sorted(
        get_entry_points_from_sdist(
            {
                "entry_points": {
                    "gui_scripts": None,
                    "console_scripts": "console_scripts=entrypoints\nfoo=bar.main",
                }
            }
        )
    ) == sorted(["console_scripts=entrypoints", "foo=bar.main"])
    assert sorted(
        get_entry_points_from_sdist(
            {
                "entry_points": {
                    "gui_scripts": "gui_scripts=entrypoints",
                    "console_scripts": None,
                }
            }
        )
    ) == sorted(["gui_scripts=entrypoints"])


def test_build_noarch_skip():
    recipe = create_python_recipe("hypothesis=5.5.2")[0]
    assert recipe["build"]["noarch"] == "python"
    assert "skip" not in recipe["build"]


@pytest.mark.github
def test_build_noarch_skip_github():
    recipe = create_python_recipe(
        "https://github.com/HypothesisWorks/hypothesis", version="5.5.2"
    )[0]
    assert recipe["build"]["noarch"] == "python"
    assert "skip" not in recipe["build"]


def test_run_requirements_sdist():
    config = Configuration(name="botocore", version="1.14.17")
    recipe = GrayskullFactory.create_recipe("pypi", config)
    assert sorted(recipe["requirements"]["run"]) == sorted(
        [
            "docutils >=0.10,<0.16",
            "jmespath >=0.7.1,<1.0.0",
            "python",
            "python-dateutil >=2.1,<3.0.0",
            "urllib3 >=1.20,<1.26",
        ]
    )


def test_format_host_requirements():
    assert sorted(format_dependencies(["setuptools>=40.0", "pkg2"], "pkg1")) == sorted(
        ["setuptools >=40.0", "pkg2"]
    )
    assert sorted(format_dependencies(["setuptools>=40.0", "pkg2"], "pkg2")) == sorted(
        ["setuptools >=40.0"]
    )
    assert sorted(format_dependencies(["setuptools >= 40.0"], "pkg")) == sorted(
        ["setuptools >=40.0"]
    )
    assert sorted(
        format_dependencies(["setuptools_scm [toml] >=3.4.1"], "pkg")
    ) == sorted(["setuptools_scm  >=3.4.1"])


def test_download_pkg_sdist(pkg_pytest):
    with open(pkg_pytest, "rb") as pkg_file:
        content = pkg_file.read()
        pkg_sha256 = hashlib.sha256(content).hexdigest()
    assert (
        pkg_sha256 == "0d5fe9189a148acc3c3eb2ac8e1ac0742cb7618c084f3d228baaec0c254b318d"
    )
    setup_cfg = get_setup_cfg(os.path.dirname(pkg_pytest))
    assert setup_cfg["name"] == "pytest"
    assert setup_cfg["python_requires"] == ">=3.5"
    assert setup_cfg["entry_points"] == {
        "console_scripts": ["pytest=pytest:main", "py.test=pytest:main"]
    }


def test_ciso_recipe():
    recipe = GrayskullFactory.create_recipe(
        "pypi", Configuration(name="ciso", version="0.1.0")
    )
    assert sorted(recipe["requirements"]["host"]) == sorted(
        ["cython", "numpy", "pip", "python"]
    )
    assert sorted(recipe["requirements"]["run"]) == sorted(
        ["cython", "python", "<{ pin_compatible('numpy') }}"]
    )
    assert recipe["test"]["commands"] == ["pip check"]
    assert recipe["test"]["requires"] == ["pip"]
    assert recipe["test"]["imports"] == ["ciso"]


@pytest.mark.serial
@pytest.mark.xfail(
    condition=(sys.platform.startswith("win")),
    reason="Test failing on windows platform",
)
def test_pymc_recipe_fortran():
    recipe = GrayskullFactory.create_recipe(
        "pypi", Configuration(name="pymc", version="2.3.6")
    )
    assert set(recipe["requirements"]["build"]) == {
        "<{ compiler('c') }}",
        "<{ compiler('fortran') }}",
    }
    assert set(recipe["requirements"]["host"]) == {"numpy", "python", "pip"}
    assert set(recipe["requirements"]["run"]) == {
        "<{ pin_compatible('numpy') }}",
        "python",
    }
    assert not recipe["build"]["noarch"]


def test_pytest_recipe_entry_points():
    recipe = create_python_recipe("pytest=5.3.5", is_strict_cf=False)[0]
    assert sorted(recipe["build"]["entry_points"]) == sorted(
        ["pytest=pytest:main", "py.test=pytest:main"]
    )
    assert recipe["about"]["license"] == "MIT"
    assert recipe["about"]["license_file"] == ["LICENSE"]
    assert "skip" in recipe["build"]
    assert recipe["build"]["skip"].inline_comment == "# [py2k]"
    assert not recipe["build"]["noarch"]
    assert sorted(recipe["test"]["commands"]) == sorted(
        ["py.test --help", "pytest --help", "pip check"]
    )


def test_cythongsl_recipe_build():
    recipe = GrayskullFactory.create_recipe(
        "pypi", Configuration(name="cythongsl", version="0.2.2")
    )

    assert recipe["requirements"]["build"] == ["<{ compiler('c') }}"]
    assert recipe["requirements"]["host"] == ["cython >=0.16", "pip", "python"]
    assert recipe["build"]["noarch"] is None
    assert recipe["build"]["number"] == 0


@pytest.mark.github
@pytest.mark.parametrize("name", ["requests", "https://github.com/psf/requests"])
def test_requests_recipe_extra_deps(capsys, name):
    CLIConfig().stdout = True
    name = parse_pkg_name_version(name)[1]
    config = Configuration(name=name, version="2.22.0")
    recipe = GrayskullFactory.create_recipe("pypi", config)
    captured_stdout = capsys.readouterr()
    assert "win-inet-pton" not in recipe["requirements"]["run"]
    assert recipe["build"]["noarch"]
    assert not recipe["build"]["skip"]
    assert f"{Fore.GREEN}{Style.BRIGHT}python" in captured_stdout.out


def test_zipp_recipe_tags_on_deps():
    config = Configuration(name="zipp", version="3.0.0")
    recipe = GrayskullFactory.create_recipe("pypi", config)
    assert recipe["build"]["noarch"]
    assert recipe["requirements"]["host"] == [
        "pip",
        "python >=3.6",
        "setuptools-scm >=3.4.1",
    ]


@pytest.mark.parametrize(
    "requires_python, expected",
    [(">=3.5, <3.8", ">=3.5,<3.8"), (">=3.7", ">=3.7"), ("~=3.6", ">=3.6")],
)
def test_generic_py_ver_to(requires_python, expected):
    config = Configuration(name="abc")
    assert generic_py_ver_to({"requires_python": requires_python}, config) == expected


def test_botocore_recipe_license_name():
    config = Configuration(name="botocore", version="1.15.8")
    recipe = GrayskullFactory.create_recipe("pypi", config)
    assert recipe["about"]["license"] == "Apache-2.0"


def test_ipytest_recipe_license():
    config = Configuration(name="ipytest", version="0.8.0")
    recipe = GrayskullFactory.create_recipe("pypi", config)
    assert recipe["about"]["license"] == "MIT"


def test_get_test_entry_points():
    assert get_test_entry_points("grayskull = grayskull.__main__:main") == [
        "grayskull --help"
    ]
    assert get_test_entry_points(
        ["pytest = py.test:main", "py.test = py.test:main"]
    ) == ["pytest --help", "py.test --help"]


def test_importlib_metadata_two_setuptools_scm():
    config = Configuration(name="importlib-metadata", version="1.5.0")
    recipe = GrayskullFactory.create_recipe("pypi", config)
    assert "setuptools-scm" in recipe["requirements"]["host"]
    assert "setuptools_scm" not in recipe["requirements"]["host"]
    assert recipe["about"]["license"] == "Apache-2.0"


def test_keyring_host_appearing_twice():
    config = Configuration(name="keyring", version="21.1.1")
    recipe = GrayskullFactory.create_recipe("pypi", config)
    assert "importlib-metadata" in recipe["requirements"]["run"]
    assert "importlib_metadata" not in recipe["requirements"]["run"]


def test_python_requires_setup_py():
    config = Configuration(name="pygments", version="2.6.1")
    recipe = GrayskullFactory.create_recipe("pypi", config)
    assert recipe["build"]["noarch"]
    assert "python >=3.5" in recipe["requirements"]["host"]
    assert "python >=3.5" in recipe["requirements"]["run"]


def test_django_rest_framework_xml_license():
    config = Configuration(name="djangorestframework-xml", version="1.4.0")
    recipe = GrayskullFactory.create_recipe("pypi", config)
    assert recipe["about"]["license"] == "BSD-3-Clause"
    assert recipe["about"]["license_file"] == ["LICENSE"]
    assert recipe["test"]["imports"][0] == "rest_framework_xml"


def test_get_test_requirements():
    config = Configuration(name="ewokscore", version="0.1.0rc5")
    recipe = GrayskullFactory.create_recipe("pypi", config)
    assert "pytest" not in recipe["test"]["requires"]
    assert "pytest --pyargs ewokscore" not in recipe["test"]["commands"]

    config = Configuration(
        name="ewokscore", version="0.1.0rc5", extras_require_test="wrongoption"
    )
    recipe = GrayskullFactory.create_recipe("pypi", config)
    assert "pytest" not in recipe["test"]["requires"]
    assert "pytest --pyargs ewokscore" not in recipe["test"]["commands"]

    # pytest dependency has no version constraints
    config = Configuration(
        name="ewokscore", version="0.1.0rc5", extras_require_test="test"
    )
    recipe = GrayskullFactory.create_recipe("pypi", config)
    assert "pytest" in recipe["test"]["requires"]
    assert "pytest --pyargs ewokscore" in recipe["test"]["commands"]

    # pytest dependency has version constraints
    config = Configuration(
        name="ewokscore", version="0.1.0rc8 ", extras_require_test="test"
    )
    recipe = GrayskullFactory.create_recipe("pypi", config)
    assert "pytest" in recipe["test"]["requires"]
    assert "pytest --pyargs ewokscore" in recipe["test"]["commands"]


def test_get_test_imports():
    assert get_test_imports({"packages": ["pkg", "pkg.mod1", "pkg.mod2"]}) == ["pkg"]
    assert get_test_imports({"packages": None}, default="pkg-mod") == ["pkg_mod"]
    assert get_test_imports({"packages": "pkg"}, default="pkg-mod") == ["pkg"]


def test_nbdime_license_type():
    config = Configuration(name="nbdime", version="2.0.0")
    recipe = GrayskullFactory.create_recipe("pypi", config)
    assert recipe["about"]["license"] == "BSD-3-Clause"
    assert "setupbase" not in recipe["requirements"]["host"]


def test_normalize_pkg_name():
    assert normalize_pkg_name("mypy-extensions") == "mypy_extensions"
    assert normalize_pkg_name("mypy_extensions") == "mypy_extensions"
    assert normalize_pkg_name("pytest") == "pytest"


def test_mypy_deps_normalization_and_entry_points():
    config = Configuration(name="mypy", version="0.770")
    recipe = GrayskullFactory.create_recipe("pypi", config)
    assert "mypy_extensions >=0.4.3,<0.5.0" in recipe["requirements"]["run"]
    assert "mypy-extensions >=0.4.3,<0.5.0" not in recipe["requirements"]["run"]
    assert "typed-ast >=1.4.0,<1.5.0" in recipe["requirements"]["run"]
    assert "typed_ast <1.5.0,>=1.4.0" not in recipe["requirements"]["run"]
    assert "typing-extensions >=3.7.4" in recipe["requirements"]["run"]
    assert "typing_extensions >=3.7.4" not in recipe["requirements"]["run"]

    assert recipe["build"]["entry_points"] == [
        "mypy=mypy.__main__:console_entry",
        "stubgen=mypy.stubgen:main",
        "stubtest=mypy.stubtest:main",
        "dmypy=mypy.dmypy.client:console_entry",
    ]


@pytest.mark.skipif(
    condition=sys.platform.startswith("win"), reason="Skipping test for win"
)
def test_panel_entry_points(tmpdir):
    config = Configuration(name="panel", version="0.9.1")
    recipe = GrayskullFactory.create_recipe("pypi", config)
    generate_recipe(recipe, config, folder_path=str(tmpdir))
    recipe_path = str(tmpdir / "panel" / "meta.yaml")
    with open(recipe_path, "r") as f:
        content = f.read()
    assert "- panel = panel.cli:main" in content


def test_deps_comments():
    config = Configuration(name="kubernetes_asyncio", version="11.2.0")
    recipe = GrayskullFactory.create_recipe("pypi", config)
    assert recipe["requirements"]["run"] == [
        "aiohttp >=2.3.10,<4.0.0",
        "certifi >=14.05.14",
        "python",
        "python-dateutil >=2.5.3",
        "pyyaml >=3.12",
        "setuptools >=21.0.0",
        "six >=1.9.0",
        "urllib3 >=1.24.2",
    ]


@pytest.mark.github
@pytest.mark.parametrize("name", ["respx=0.10.1", "https://github.com/lundberg/respx"])
def test_keep_filename_license(name):
    recipe = create_python_recipe(name)[0]
    assert recipe["about"]["license_file"] == ["LICENSE.md"]


def test_platform_system_selector():
    assert parse_extra_metadata_to_selector("platform_system", "==", "Windows") == "win"
    assert (
        parse_extra_metadata_to_selector("platform_system", "!=", "Windows")
        == "not win"
    )


def test_tzdata_without_setup_py():
    recipe = create_python_recipe("tzdata=2020.1")[0]
    assert recipe["build"]["noarch"] == "python"
    assert recipe["about"]["home"] == "https://github.com/python/tzdata"


def test_multiples_exit_setup():
    """Bug fix #146"""
    assert create_python_recipe("pyproj=2.6.1")[0]


def test_sequence_inside_another_in_dependencies():
    recipe = create_python_recipe("unittest2=1.1.0", is_strict_cf=True)[0]
    assert recipe["requirements"]["host"] == [
        "argparse",
        "pip",
        "python >=3.6",
        "six >=1.4",
        "traceback2",
    ]
    assert recipe["requirements"]["run"] == [
        "argparse",
        "python >=3.6",
        "six >=1.4",
        "traceback2",
    ]


def test_recipe_with_just_py_modules():
    recipe = create_python_recipe("python-markdown-math=0.7")[0]
    assert recipe["test"]["imports"] == ["mdx_math"]


def test_recipe_extension():
    recipe = create_python_recipe("azure-identity=1.3.1")[0]
    assert (
        recipe["source"]["url"]
        == "https://pypi.io/packages/source/{{ name[0] }}/{{ name }}/"
        "azure-identity-{{ version }}.zip"
    )


def test_get_url_filename():
    assert get_url_filename({}) == "{{ name }}-{{ version }}.tar.gz"
    assert get_url_filename({}, "default") == "default"
    assert (
        get_url_filename({"urls": [{"packagetype": "nothing"}]})
        == "{{ name }}-{{ version }}.tar.gz"
    )
    assert (
        get_url_filename({"urls": [{"packagetype": "nothing"}]}, "default") == "default"
    )
    assert (
        get_url_filename(
            {
                "info": {"version": "1.2.3"},
                "urls": [{"packagetype": "sdist", "filename": "foo_file-1.2.3.zip"}],
            }
        )
        == "foo_file-{{ version }}.zip"
    )


def test_clean_deps_for_conda_forge():
    assert clean_deps_for_conda_forge(["deps1", "deps2  # [py34]"], PyVer(3, 6)) == [
        "deps1"
    ]
    assert clean_deps_for_conda_forge(["deps1", "deps2  # [py<34]"], PyVer(3, 6)) == [
        "deps1"
    ]
    assert clean_deps_for_conda_forge(["deps1", "deps2  # [py<38]"], PyVer(3, 6)) == [
        "deps1",
        "deps2  # [py<38]",
    ]


def test_empty_entry_points():
    recipe = create_python_recipe("modulegraph=0.18")[0]
    assert recipe["build"]["entry_points"] == [
        "modulegraph = modulegraph.__main__:main"
    ]


def test_noarch_metadata():
    recipe = create_python_recipe("policy_sentry=0.11.16")[0]
    assert recipe["build"]["noarch"] == "python"


def test_arch_metadata():
    recipe = create_python_recipe("remove_dagmc_tags=0.0.5")[0]
    assert "noarch" not in recipe["build"]


def test_entry_points_is_list_of_str():
    """Test to verify that whether console_scripts is a list of strings,
    a multiline string, or a list of empty lists; entry_points is always a list"""
    sdist_metadata = {
        "entry_points": {
            "console_scripts": [
                "ptpython = ptpython.entry_points.run_ptpython:run",
                "ptipython = ptpython.entry_points.run_ptipython:run",
                "ptpython3 = ptpython.entry_points.run_ptpython:run",
                "ptpython3.9 = ptpython.entry_points.run_ptpython:run",
                "ptipython3 = ptpython.entry_points.run_ptipython:run",
                "ptipython3.9 = ptpython.entry_points.run_ptipython:run",
            ]
        },
    }
    assert isinstance(get_entry_points_from_sdist(sdist_metadata), list)
    sdist_metadata = {
        "entry_points": {
            "console_scripts": """
                ptpython = ptpython.entry_points.run_ptpython:run
                ptipython = ptpython.entry_points.run_ptipython:run
                ptpython3 = ptpython.entry_points.run_ptpython:run
                ptpython3.9 = ptpython.entry_points.run_ptpython:run
                ptipython3 = ptpython.entry_points.run_ptipython:run
                ptipython3.9 = ptpython.entry_points.run_ptipython:run
                """
        },
    }
    assert isinstance(get_entry_points_from_sdist(sdist_metadata), list)
    sdist_metadata = {
        "entry_points": {"console_scripts": [[]]},
    }
    assert isinstance(get_entry_points_from_sdist(sdist_metadata), list)


def test_replace_slash_in_imports():
    recipe = create_python_recipe("asgi-lifespan=1.0.1")[0]
    assert ["asgi_lifespan"] == recipe["test"]["imports"]


def test_add_python_min_to_strict_conda_forge():
    recipe = create_python_recipe("dgllife=0.2.8", is_strict_cf=True)[0]
    assert recipe["build"]["noarch"] == "python"
    assert recipe["requirements"]["host"][1] == "python >=3.6"
    assert "python >=3.6" in recipe["requirements"]["run"]


def test_get_test_imports_clean_modules():
    assert get_test_imports(
        {
            "packages": [
                "_pytest",
                "tests",
                "test",
                "_pytest._code",
                "_pytest._io",
                "_pytest.assertion",
                "_pytest.config",
                "_pytest.mark",
                "pytest",
                "pytest.foo",
                "zar",
            ]
        }
    ) == ["pytest", "zar"]
    assert get_test_imports(
        {
            "packages": [
                "_pytest",
                "_pytest._code",
                "_pytest._io",
                "_pytest.assertion",
                "_pytest.config",
                "_pytest.mark",
            ]
        }
    ) == ["_pytest", "_pytest._code"]


def test_ensure_pep440():
    assert ensure_pep440("pytest ~=5.3.2") == "pytest >=5.3.2,<5.4.dev0"


def test_pep440_recipe():
    recipe = create_python_recipe("codalab=0.5.26", is_strict_cf=False)[0]
    assert recipe["requirements"]["host"] == ["pip", "python >=3.6"]


def test_pep440_in_recipe_pypi():
    recipe = create_python_recipe("kedro=0.17.6", is_strict_cf=False)[0]
    assert sorted(recipe["requirements"]["run"])[0] == "anyconfig >=0.10.0,<0.11.dev0"


def test_create_recipe_from_local_sdist(pkg_pytest):
    recipe = create_python_recipe(pkg_pytest, from_local_sdist=True)[0]
    assert recipe["source"]["url"] == f"file://{pkg_pytest}"
    assert recipe["about"]["home"] == "https://docs.pytest.org/en/latest/"
    assert recipe["about"]["summary"] == "pytest: simple powerful testing with Python"
    assert recipe["about"]["license"] == "MIT"
    assert recipe["about"]["license_file"] == ["LICENSE"]


def test_400_for_python_selector():
    recipe = create_python_recipe("pyquil", version="3.0.1")[0]
    assert recipe["build"]["skip"].selector == "py>=400 or py2k"


def test_update_requirements_with_pin_mixed_numpy_pin_compatible():
    requirements = {
        "build": ["<{ compiler('c') }}"],
        "host": ["cython", "numpy", "pip", "python"],
        "run": ["numpy >=1.19.1,<2.0.0", "pyparsing >=2.4.7, <3.0.0", "python"],
    }
    update_requirements_with_pin(requirements)
    assert "<{ pin_compatible('numpy') }}" in requirements["run"]
    assert "numpy" not in requirements["run"]


def test_notice_file():
    recipe, _ = create_python_recipe(
        "apache-airflow-providers-databricks", version="3.1.0"
    )
    assert set(recipe["about"]["license_file"]) == {"NOTICE", "LICENSE"}
    assert recipe["about"]["license"] == "Apache-2.0"


def test_notice_file_different_licence():
    with patch(
        "grayskull.license.discovery.get_license_type",
        side_effect=["Apache-2.0", "MIT"],
    ):
        recipe, _ = create_python_recipe(
            "apache-airflow-providers-databricks", version="3.1.0"
        )
    assert set(recipe["about"]["license_file"]) == {"NOTICE", "LICENSE"}
    assert recipe["about"]["license"] in ["MIT & Apache-2.0", "Apache-2.0 & MIT"]


def test_console_script_toml_format():
    recipe, _ = create_python_recipe("consolemd", version="0.5.1")
    assert recipe["build"]["entry_points"] == ["consolemd = consolemd.cli:cli"]


def test_section_order():
    recipe, _ = create_python_recipe("requests", version="2.27.1")
    assert (
        "package",
        "source",
        "build",
        "requirements",
        "test",
        "about",
        "extra",
    ) == tuple(recipe.keys())


def test_no_sdist_pkg_pypi():
    with pytest.raises(
        AttributeError, match="There is no sdist package on pypi for arn"
    ):
        recipe, _ = create_python_recipe("arn", version="0.1.5")


def test_remove_selectors_pkgs_if_needed():
    assert remove_selectors_pkgs_if_needed(
        [
            "import_metadata >1.0  # [py>3]",
            "pywin32  # [win]",
            "requests >=2.0  # [unix]",
        ]
    ) == ["import_metadata >1.0", "pywin32", "requests >=2.0  # [unix]"]


def test_remove_selectors_pkgs_if_needed_with_recipe():
    recipe, _ = create_python_recipe("transformers", is_strict_cf=True, version="4.3.3")
    assert set(recipe["requirements"]["run"]).issubset(
        {
            "dataclasses",
            "filelock",
            "importlib-metadata",
            "numpy >=1.17",
            "packaging",
            "python",
            "regex !=2019.12.17",
            "requests",
            "sacremoses",
            "tokenizers <0.11,>=0.10.1",
            "tokenizers >=0.10.1,<0.11",
            "tqdm >=4.27",
        }
    )


def test_noarch_python_min_constrain():
    recipe, _ = create_python_recipe("humre", is_strict_cf=True, version="0.1.1")
    assert recipe["requirements"]["run"] == ["python >=3.6"]


def test_cpp_language_extra():
    recipe, _ = create_python_recipe("xbcausalforest", version="0.1.3")
    assert set(recipe["requirements"]["build"]) == {
        "<{ compiler('cxx') }}",
        "<{ compiler('c') }}",
    }



import ast
import hashlib
import logging
import os
import re
from collections import namedtuple
from difflib import SequenceMatcher
from functools import lru_cache
from glob import glob
from pathlib import Path
from shutil import copyfile
from typing import List, Optional, Union

from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap
from souschef.section import Section

log = logging.getLogger(__name__)

PyVer = namedtuple("PyVer", ["major", "minor"])
yaml = YAML(typ="jinja2")
yaml.indent(mapping=2, sequence=4, offset=2)
yaml.width = 600


@lru_cache(maxsize=10)
def get_std_modules() -> List:
    from stdlib_list import stdlib_list

    all_libs = set()
    for py_ver in (
        "2.7",
        "3.6",
        "3.7",
        "3.8",
    ):
        all_libs.update(stdlib_list(py_ver))
    return list(all_libs)


def get_all_modules_imported_script(script_file: str) -> set:
    modules = set()

    def visit_Import(node):
        for name in node.names:
            if name.name:
                modules.add(name.name.split(".")[0])

    def visit_ImportFrom(node):
        # if node.module is missing it's a "from . import ..." statement
        # if level > 0 it's a "from .submodule import ..." statement
        if node.module is not None and node.level == 0 and node.module:
            modules.add(node.module.split(".")[0])

    node_iter = ast.NodeVisitor()
    node_iter.visit_Import = visit_Import
    node_iter.visit_ImportFrom = visit_ImportFrom
    with open(script_file, "r") as f:
        node_iter.visit(ast.parse(f.read()))
    return modules


def get_vendored_dependencies(script_file: str) -> List:
    """Get all third part dependencies which are being in use in the setup.py

    :param script_file: Path to the setup.py
    :return: List with all vendored dependencies
    """
    all_std_modules = get_std_modules()
    all_modules_used = get_all_modules_imported_script(script_file)
    local_modules = get_local_modules(os.path.dirname(script_file))
    return [
        dep.lower()
        for dep in all_modules_used
        if dep not in local_modules and dep not in all_std_modules
    ]


@lru_cache(maxsize=20)
def get_local_modules(sdist_folder: str) -> List:
    result = []
    for py_file in glob(f"{sdist_folder}/*.py"):
        py_file = os.path.basename(py_file)
        if py_file == "setup.py":
            continue
        result.append(os.path.splitext(py_file)[0])
    return result


def origin_is_github(name_or_url: str) -> bool:
    return (
        name_or_url.startswith(("http://", "https://")) and "github.com" in name_or_url
    )  # lgtm [py/incomplete-url-substring-sanitization]


def origin_is_local_sdist(name: str) -> bool:
    """Return True if the given local file can be a sdist"""
    # Available formats according to
    # https://docs.python.org/3/distutils/sourcedist.html#creating-a-source-distribution
    return (
        name.endswith((".tar.gz", ".tar.bz2", ".tar.xz", ".tar.Z", ".tar", ".zip"))
        and Path(name).is_file()
    )


def sha256_checksum(filename, block_size=65536):
    sha256 = hashlib.sha256()
    with open(filename, "rb") as f:
        for block in iter(lambda: f.read(block_size), b""):
            sha256.update(block)
    return sha256.hexdigest()


def string_similarity(a, b):
    return SequenceMatcher(None, a, b).ratio()


def rm_duplicated_deps(all_requirements: Union[list, set, None]) -> Optional[list]:
    if not all_requirements:
        return None
    new_value = []
    for dep in all_requirements:
        if (
            dep in new_value
            or dep.replace("-", "_") in new_value
            or dep.replace("_", "-") in new_value
        ):
            continue
        new_value.append(dep)
    return new_value


def format_dependencies(all_dependencies: List, name: str) -> List:
    """Just format the given dependency to a string which is valid for the
    recipe

    :param all_dependencies: list of dependencies
    :param name: package name
    :return: list of dependencies formatted
    """
    formatted_dependencies = []
    re_deps = re.compile(r"^\s*([\.a-zA-Z0-9_-]+)\s*(.*)\s*$", re.MULTILINE | re.DOTALL)
    re_remove_space = re.compile(r"([<>!=]+)\s+")
    re_remove_tags = re.compile(r"\s*(\[.*\])", re.DOTALL)
    re_remove_comments = re.compile(r"\s+#.*", re.DOTALL)
    for req in all_dependencies:
        match_req = re_deps.match(req)
        deps_name = req
        if deps_name.replace("-", "_") == name.replace("-", "_"):
            continue
        if match_req:
            match_req = match_req.groups()
            deps_name = match_req[0]
            if len(match_req) > 1:
                deps_name = " ".join(match_req)
        deps_name = re_remove_space.sub(r"\1", deps_name.strip())
        deps_name = re_remove_tags.sub(r" ", deps_name.strip())
        deps_name = re_remove_comments.sub("", deps_name)
        formatted_dependencies.append(deps_name.strip())
    return formatted_dependencies


def generate_recipe(
    recipe,
    config,
    folder_path: Union[str, Path] = ".",
):
    """Write the recipe in a location. It will create a folder with the
    package name and the recipe will be there.

    :param folder_path: Path to the folder
    """
    pkg_name = config.name
    if origin_is_github(pkg_name):
        pkg_name = pkg_name.split("/")[-1]
    if Path(folder_path).is_file():
        folder_path = Path(folder_path)
        recipe_path = folder_path
        recipe_folder = folder_path.parent
    else:
        recipe_dir = Path(folder_path) / pkg_name
        logging.debug(f"Generating recipe on: {recipe_dir}")
        if not recipe_dir.is_dir():
            recipe_dir.mkdir()
        recipe_path = recipe_dir / "meta.yaml"
        recipe_folder = recipe_dir
        add_new_lines_after_section(recipe.yaml)

    clean_yaml(recipe)
    recipe.save(recipe_path)
    for file_to_recipe in config.files_to_copy:
        name = file_to_recipe.split(os.path.sep)[-1]
        if os.path.isfile(file_to_recipe):
            copyfile(file_to_recipe, os.path.join(recipe_folder, name))


def get_clean_yaml(recipe_yaml: CommentedMap) -> CommentedMap:
    clean_yaml(recipe_yaml)
    return add_new_lines_after_section(recipe_yaml)


def add_new_lines_after_section(recipe_yaml: CommentedMap) -> CommentedMap:
    for section in recipe_yaml.keys():
        if section == "package":
            recipe_yaml.yaml_set_comment_before_after_key(section, "\n\n")
        else:
            recipe_yaml.yaml_set_comment_before_after_key(section, "\n")
    return recipe_yaml


def clean_yaml(recipe):
    for key, yaml_obj in _clean_yaml(recipe):
        del yaml_obj[key]


def _clean_yaml(recipe, all_obj_to_delete=None):
    all_obj_to_delete = all_obj_to_delete or []
    for key, value in recipe.items():
        value_to_delete = []
        if not isinstance(value, bool) and not value:
            value_to_delete = [(key, recipe)]
        elif isinstance(value, Section):
            value_to_delete = _clean_yaml(value)
        all_obj_to_delete.extend(value_to_delete)
    return all_obj_to_delete


def merge_list_item(destination: dict, add: dict, key: str) -> None:
    lst = set(destination.get(key, []))
    lst |= set(add.get(key, []))
    if lst:
        destination[key] = list(lst)


def merge_dict_of_lists_item(destination: dict, add: dict, key: str) -> None:
    sub_destination = destination.get(key, {})
    sub_add = add.get(key, {})
    for sub_key in set(sub_destination) | set(sub_add):
        merge_list_item(sub_destination, sub_add, sub_key)
    if sub_destination:
        destination[key] = sub_destination



import logging
import os
import re
import shutil
import sys
from contextlib import contextmanager
from copy import deepcopy
from distutils import core
from pathlib import Path
from subprocess import check_output
from tempfile import mkdtemp
from typing import Dict, List, Optional, Tuple, Union
from urllib.parse import urlparse

import requests
import tomli
from colorama import Fore, Style
from packaging.utils import canonicalize_version
from packaging.version import Version
from pkginfo import UnpackedSDist

from grayskull.cli.stdout import manage_progressbar, print_msg
from grayskull.config import Configuration
from grayskull.license.discovery import ShortLicense, search_license_file
from grayskull.utils import (
    PyVer,
    get_vendored_dependencies,
    merge_dict_of_lists_item,
    merge_list_item,
    origin_is_github,
    sha256_checksum,
)

log = logging.getLogger(__name__)
RE_DEPS_NAME = re.compile(r"^\s*([\.a-zA-Z0-9_-]+)", re.MULTILINE)
PIN_PKG_COMPILER = {"numpy": "<{ pin_compatible('numpy') }}"}


def search_setup_root(path_folder: Union[Path, str]) -> Path:
    if setup_py := list(Path(path_folder).rglob("setup.py")):
        return setup_py[0]
    if setup_cfg := list(Path(path_folder).rglob("setup.cfg")):
        return setup_cfg[0]
    if pyproject_toml := list(Path(path_folder).rglob("pyproject.toml")):
        return pyproject_toml[0]


def clean_deps_for_conda_forge(list_deps: List, py_ver_min: PyVer) -> List:
    """Remove dependencies which conda-forge is not supporting anymore.
    For example Python 2.7, Python version less than 3.6"""
    re_delimiter = re.compile(r"#\s+\[py\s*(?:([<>=!]+))?\s*(\d+)\]\s*$", re.DOTALL)
    result_deps = []
    for dependency in list_deps:
        match_del = re_delimiter.search(dependency)
        if match_del is None:
            result_deps.append(dependency)
            continue

        match_del = match_del.groups()
        if not match_del[0]:
            match_del = ("==", match_del[1])
        major = int(match_del[1][0])
        minor = int(match_del[1][1:].replace("k", "0") or 0)
        current_py = PyVer(major=major, minor=minor)
        log.debug(f"Evaluating: {py_ver_min}{match_del}{current_py} -- {dependency}")
        if eval(f"py_ver_min{match_del[0]}current_py"):
            result_deps.append(dependency)
    return result_deps


def pkg_name_from_sdist_url(sdist_url: str):
    """This method extracts and returns the name of the package from the sdist
    url."""
    if origin_is_github(sdist_url):
        return sdist_url.split("/")[-3] + ".tar.gz"
    else:
        return sdist_url.split("/")[-1]


def parse_extra_metadata_to_selector(option: str, operation: str, value: str) -> str:
    """Method tries to convert the extra metadata received into selectors

    :param option: option (extra, python_version, sys_platform)
    :param operation: (>, >=, <, <=, ==, !=)
    :param value: value after the operation
    :return: selector
    """
    if option == "extra":
        return ""
    if option == "python_version":
        value = value.split(".")
        value = "".join(value[:2])
        return f"py{operation}{value}"
    if option == "sys_platform":
        value = re.sub(r"[^a-zA-Z]+", "", value)
        if operation == "!=":
            return f"not {value.lower()}"
        return value.lower()
    if option == "platform_system":
        replace_val = {"windows": "win", "linux": "linux", "darwin": "osx"}
        value_lower = value.lower().strip()
        if value_lower in replace_val:
            value_lower = replace_val[value_lower]
        if operation == "!=":
            return f"not {value_lower}"
        return value_lower


def get_extra_from_requires_dist(string_parse: str) -> Union[List]:
    """Receives the extra metadata e parse it to get the option, operation
    and value.

    :param string_parse: metadata extra
    :return: return the option , operation and value of the extra metadata
    """
    return re.findall(
        r"(?:(\())?\s*([\.a-zA-Z0-9-_]+)\s*([=!<>]+)\s*[\'\"]*"
        r"([\.a-zA-Z0-9-_]+)[\'\"]*\s*(?:(\)))?\s*(?:(and|or))?",
        string_parse,
    )


def get_name_version_from_requires_dist(string_parse: str) -> Tuple[str, str]:
    """Extract the name and the version from `requires_dist` present in
    PyPi`s metadata

    :param string_parse: requires_dist value from PyPi metadata
    :return: Name and version of a package
    """
    pkg = re.match(r"^\s*([^\s]+)\s*([\(]*.*[\)]*)?\s*", string_parse, re.DOTALL)
    pkg_name = pkg.group(1).strip()
    version = ""
    if len(pkg.groups()) > 1 and pkg.group(2):
        version = " " + pkg.group(2).strip()
    return pkg_name.strip(), re.sub(r"[\(\)]", "", version).strip()


def generic_py_ver_to(
    metadata: dict, config, is_selector: bool = False
) -> Optional[str]:  # sourcery no-metrics
    """Generic function which abstract the parse of the requires_python
    present in the PyPi metadata. Basically it can generate the selectors
    for Python or the constrained version if it is a `noarch: python` python package"""
    if not metadata.get("requires_python"):
        return None
    req_python = re.findall(
        r"([~><=!]+)\s*(\d+)(?:\.(\d+))?",
        metadata["requires_python"],
    )
    if not req_python:
        return None

    py_ver_enabled = config.get_py_version_available(req_python)
    small_py3_version = config.get_oldest_py3_version(list(py_ver_enabled.keys()))
    all_py = list(py_ver_enabled.values())
    if all(all_py):
        return None
    if all(all_py if config.is_strict_cf else all_py[1:]):
        if is_selector:
            return None if config.is_strict_cf else "# [py2k]"
        else:
            return f">={small_py3_version.major}.{small_py3_version.minor}"
    if py_ver_enabled.get(PyVer(2, 7)) and any(all_py[1:]) is False:
        return "# [py3k]" if is_selector else "<3.0"

    for pos, py_ver in enumerate(py_ver_enabled):
        if py_ver == PyVer(2, 7):
            continue
        if all(all_py[pos:]) and any(all_py[:pos]) is False:
            if is_selector:
                minor = f"{py_ver.minor:02d}" if py_ver.major >= 4 else py_ver.minor
                return f"# [py<{py_ver.major}{minor}]"
            else:
                return f">={py_ver.major}.{py_ver.minor}"
        elif any(all_py[pos:]) is False:
            if is_selector:
                py2k = ""
                if not config.is_strict_cf and not all_py[0]:
                    py2k = " or py2k"
                minor = f"{py_ver.minor:02d}" if py_ver.major >= 4 else py_ver.minor
                return f"# [py>={py_ver.major}{minor}{py2k}]"
            else:
                py2 = ""
                if not all_py[0]:
                    py2 = f">={small_py3_version.major}.{small_py3_version.minor},"
                return f"{py2}<{py_ver.major}.{py_ver.minor}"

    all_selector = get_py_multiple_selectors(
        py_ver_enabled, is_selector=is_selector, config=config
    )
    if all_selector:
        return (
            "# [{}]".format(" or ".join(all_selector))
            if is_selector
            else ",".join(all_selector)
        )
    return None


def install_deps_if_necessary(setup_path: str, data_dist: dict, pip_dir: str):
    """Install missing dependencies to run the setup.py

    :param setup_path: path to the setup.py
    :param data_dist: metadata
    :param pip_dir: path where the missing packages will be downloaded
    """
    all_setup_deps = get_vendored_dependencies(setup_path)
    for dep in all_setup_deps:
        pip_install_dep(data_dist, dep, pip_dir)


def pip_install_dep(data_dist: dict, dep_name: str, pip_dir: str):
    """Install dependency using `pip`

    :param data_dist: sdist metadata
    :param dep_name: Package name which will be installed
    :param pip_dir: Path to the folder where `pip` will let the packages
    """
    if not data_dist.get("setup_requires"):
        data_dist["setup_requires"] = []
    if dep_name == "pkg_resources":
        dep_name = "setuptools"
    try:
        check_output(
            [
                sys.executable,
                "-m",
                "pip",
                "install",
                dep_name,
                f"--target={pip_dir}",
            ]
        )
    except Exception as err:
        log.error(
            f"It was not possible to install {dep_name}.\n"
            f"Command: pip install {dep_name} --target={pip_dir}.\n"
            f"Error: {err}"
        )
    else:
        if (
            dep_name.lower() not in data_dist["setup_requires"]
            and dep_name.lower() != "setuptools"
        ):
            data_dist["setup_requires"].append(dep_name.lower())


def merge_sdist_metadata(setup_py: dict, setup_cfg: dict) -> dict:
    """This method will merge the metadata present in the setup.py and
    setup.cfg. It is an auxiliary method.

    :param setup_py: Metadata from setup.py
    :param setup_cfg: Metadata from setup.cfg
    :return: Return the merged data from setup.py and setup.cfg
    """
    result = deepcopy(setup_py)
    for key, value in setup_cfg.items():
        if key not in result:
            result[key] = value

    merge_list_item(result, setup_cfg, "install_requires")
    merge_list_item(result, setup_cfg, "setup_requires")
    merge_list_item(result, setup_cfg, "compilers")
    merge_dict_of_lists_item(result, setup_cfg, "extras_require")

    if "setuptools-scm" in result.get("setup_requires", []):
        result["setup_requires"].remove("setuptools-scm")

    return result


def get_setup_cfg(source_path: str) -> dict:
    """Method responsible to extract the setup.cfg metadata

    :param source_path: Path to the folder where is located the sdist
     files unpacked
    :return: Metadata of setup.cfg
    """
    try:
        from setuptools.config.setupcfg import read_configuration
    except ImportError:
        from setuptools.config import read_configuration

    log.debug(f"Started setup.cfg from {source_path}")
    print_msg("Recovering metadata from setup.cfg")
    path_setup_cfg = list(Path(source_path).rglob("setup.cfg"))
    if not path_setup_cfg:
        return {}
    path_setup_cfg = path_setup_cfg[0]

    setup_cfg = read_configuration(str(path_setup_cfg))
    setup_cfg = dict(setup_cfg)
    if setup_cfg.get("options", {}).get("python_requires"):
        setup_cfg["options"]["python_requires"] = ensure_pep440(
            str(setup_cfg["options"]["python_requires"])
        )
    result = {}
    result.update(setup_cfg.get("options", {}))
    result.update(setup_cfg.get("metadata", {}))
    if result.get("build_ext"):
        result["compilers"] = ["c"]
    log.debug(f"Data recovered from setup.cfg: {result}")
    return result


@contextmanager
def injection_distutils(folder: str) -> dict:
    """This is a bit of "dark magic", please don't do it at home.
    It is injecting code in the distutils.core.setup and replacing the
    setup function by the inner function __fake_distutils_setup.
    This method is a contextmanager, after leaving the context it will return
    with the normal implementation of the distutils.core.setup.
    This method is necessary for two reasons:
    -pypi metadata for some packages might be absent from the pypi API.
    -pypi metadata, when present, might be missing some information.

    :param folder: Path to the folder where the sdist package was extracted
    :yield: return the metadata from sdist
    """
    from distutils import core

    setup_core_original = core.setup
    old_dir = os.getcwd()
    path_setup = search_setup_root(folder)
    os.chdir(os.path.dirname(str(path_setup)))

    data_dist = {}

    def __fake_distutils_setup(*args, **kwargs):
        if not isinstance(kwargs, dict) or not kwargs:
            return

        def _fix_list_requirements(key_deps: str) -> List:
            """Fix when dependencies have lists inside of another sequence"""
            val_deps = kwargs.get(key_deps)
            if not val_deps:
                return val_deps
            list_req = []
            if isinstance(val_deps, str):
                val_deps = [val_deps]
            for val in val_deps:
                if isinstance(val, (tuple, list)):
                    list_req.extend(list(map(str, val)))
                else:
                    list_req.append(str(val))
            return list_req

        if "setup_requires" in kwargs:
            kwargs["setup_requires"] = _fix_list_requirements("setup_requires")
        if "install_requires" in kwargs:
            kwargs["install_requires"] = _fix_list_requirements("install_requires")

        data_dist.update(kwargs)
        if not data_dist.get("setup_requires"):
            data_dist["setup_requires"] = []
        data_dist["setup_requires"] += (
            kwargs.get("setup_requires") if kwargs.get("setup_requires") else []
        )

        if "use_scm_version" in data_dist and kwargs["use_scm_version"]:
            log.debug("setuptools_scm found on setup.py")
            if "setuptools_scm" not in data_dist["setup_requires"]:
                data_dist["setup_requires"] += ["setuptools_scm"]
            if "setuptools-scm" in data_dist["setup_requires"]:
                data_dist["setup_requires"].remove("setuptools-scm")

        if kwargs.get("ext_modules", None):
            compilers = {"c"}
            for module in kwargs.get("ext_modules", []):
                if (getattr(module, "language", "") or "").lower() in {"c++", "cpp"}:
                    compilers.add("cxx")
                elif (
                    hasattr(module, "has_f2py_sources") and module.has_f2py_sources()
                ) or (getattr(module, "language", "") or "").lower() in (
                    "fortran",
                    "f77",
                    "f90",
                ):
                    compilers.add("fortran")
            data_dist["compilers"] = list(compilers)
        log.debug(f"Injection distutils all arguments: {kwargs}")
        if data_dist.get("run_py", False):
            del data_dist["run_py"]
            return
        setup_core_original(*args, **kwargs)

    try:
        core.setup = __fake_distutils_setup
        path_setup = str(path_setup)
        print_msg("Executing injected distutils...")
        __run_setup_py(path_setup, data_dist)
        if not data_dist or not data_dist.get("install_requires", None):
            print_msg(
                "No data was recovered from setup.py."
                " Forcing to execute the setup.py as script"
            )
            __run_setup_py(path_setup, data_dist, run_py=True)
        yield data_dist
    except BaseException as err:
        log.debug(f"Exception occurred when executing sdist injection: {err}")
        yield data_dist
    finally:
        core.setup = setup_core_original
        os.chdir(old_dir)


def __run_setup_py(path_setup: str, data_dist: dict, run_py=False, deps_installed=None):
    """Method responsible to run the setup.py

    :param path_setup: full path to the setup.py
    :param data_dist: metadata
    :param run_py: If it should run the setup.py with run_py, otherwise it will run
    invoking the distutils directly
    """
    deps_installed = deps_installed or []
    original_path = deepcopy(sys.path)
    pip_dir = mkdtemp(prefix="pip-dir-")
    if not os.path.exists(pip_dir):
        os.mkdir(pip_dir)
    if os.path.dirname(path_setup) not in sys.path:
        sys.path.append(os.path.dirname(path_setup))
        sys.path.append(pip_dir)
    install_deps_if_necessary(path_setup, data_dist, pip_dir)
    try:
        if run_py:
            import runpy

            data_dist["run_py"] = True
            runpy.run_path(path_setup, run_name="__main__")
        else:
            core.run_setup(path_setup, script_args=["install", f"--target={pip_dir}"])
    except ModuleNotFoundError as err:
        log.debug(
            f"When executing setup.py did not find the module: {err.name}."
            f" Exception: {err}"
        )
        dep_install = err.name
        if dep_install in deps_installed:
            dep_install = dep_install.split(".")[0]
        if dep_install not in deps_installed:
            deps_installed.append(dep_install)
            pip_install_dep(data_dist, dep_install, pip_dir)
            __run_setup_py(path_setup, data_dist, run_py, deps_installed=deps_installed)
    except Exception as err:
        log.debug(f"Exception when executing setup.py as script: {err}")
    data_dist.update(
        merge_sdist_metadata(data_dist, get_setup_cfg(os.path.dirname(str(path_setup))))
    )
    log.debug(f"Data recovered from setup.py: {data_dist}")
    if os.path.exists(pip_dir):
        shutil.rmtree(pip_dir)
    sys.path = original_path


def get_compilers(
    requires_dist: List, sdist_metadata: dict, config: Configuration
) -> List:
    """Return which compilers are necessary"""
    compilers = set(sdist_metadata.get("compilers", []))
    for pkg in requires_dist:
        pkg = RE_DEPS_NAME.match(pkg).group(0)
        pkg = pkg.lower().strip()
        if pkg.strip() in config.pkg_need_c_compiler:
            compilers.add("c")
        if pkg.strip() in config.pkg_need_cxx_compiler:
            compilers.add("cxx")
    return list(compilers)


def get_py_multiple_selectors(
    selectors: Dict[PyVer, bool],
    config: Configuration,
    is_selector: bool = False,
) -> List:
    """Get python selectors available.

    :param selectors: Dict with the Python version and if it is selected
    :param is_selector: if it needs to convert to selector or constrain python
    :param config: Configuration object
    :return: list with all selectors or constrained python
    """
    all_selector = []
    if not config.is_strict_cf and selectors[PyVer(2, 7)] is False:
        all_selector += (
            ["py2k"]
            if is_selector
            else config.get_oldest_py3_version(list(selectors.keys()))
        )
    for py_ver, is_enabled in selectors.items():
        if (not config.is_strict_cf and py_ver == PyVer(2, 7)) or is_enabled:
            continue
        all_selector += (
            [f"py=={py_ver.major}{py_ver.minor}"]
            if is_selector
            else [f"!={py_ver.major}.{py_ver.minor}"]
        )
    return all_selector


def py_version_to_selector(pypi_metadata: dict, config) -> Optional[str]:
    return generic_py_ver_to(pypi_metadata, is_selector=True, config=config)


def py_version_to_limit_python(pypi_metadata: dict, config=None) -> Optional[str]:
    config = config or Configuration(pypi_metadata["name"])
    result = generic_py_ver_to(pypi_metadata, is_selector=False, config=config)
    if not result and config.is_strict_cf:
        result = (
            f">={config.py_cf_supported[0].major}.{config.py_cf_supported[0].minor}"
        )
    return result


def update_requirements_with_pin(requirements: dict):
    """Get a dict with the `host`, `run` and `build` in it and replace
    if necessary the run requirements with the appropriated pin.

    :param requirements: Dict with the requirements in it
    """

    def is_compiler_present() -> bool:
        if "build" not in requirements:
            return False
        re_compiler = re.compile(
            r"^\s*[<{]\{\s*compiler\(['\"]\w+['\"]\)\s*\}\}\s*$", re.MULTILINE
        )
        return any(re_compiler.match(build) for build in requirements["build"])

    if not is_compiler_present():
        return

    def clean_list_pkg(pkg, list_pkgs):
        return [p for p in list_pkgs if pkg != p.strip().split(" ", 1)[0]]

    for pkg in requirements["host"]:
        pkg_name = RE_DEPS_NAME.match(pkg).group(0)
        if pkg_name in PIN_PKG_COMPILER.keys():
            requirements["run"] = clean_list_pkg(pkg_name, requirements["run"])
            requirements["run"].append(PIN_PKG_COMPILER[pkg_name])


def discover_license(metadata: dict) -> List[ShortLicense]:
    """Based on the metadata this method will try to discover what is the
    right license for the package

    :param metadata: metadata
    :return: Return an object which contains relevant information regarding
    the license.
    """
    git_url = metadata.get("dev_url")
    if not git_url and urlparse(metadata.get("project_url", "")).netloc == "github.com":
        git_url = metadata.get("project_url")
    # "url" is always present but sometimes set to None
    if not git_url and urlparse((metadata.get("url") or "")).netloc == "github.com":
        git_url = metadata.get("url")

    return search_license_file(
        metadata.get("sdist_path"),
        git_url,
        metadata.get("version"),
        license_name_metadata=metadata.get("license"),
    )


def get_test_entry_points(entry_points: Union[List, str]) -> List:
    if entry_points and isinstance(entry_points, str):
        entry_points = [entry_points]
    return [f"{ep.split('=')[0].strip()} --help" for ep in entry_points]


def get_test_imports(metadata: dict, default: Optional[str] = None) -> List:
    if default:
        default = default.replace("-", "_")
    if "packages" not in metadata or not metadata["packages"]:
        return [default]
    meta_pkg = metadata["packages"]
    if isinstance(meta_pkg, str):
        meta_pkg = [metadata["packages"]]
    result = []
    for module in sorted(meta_pkg):
        if "/" in module or "." in module or module.startswith("_"):
            continue
        if module in ["test", "tests"]:
            log.warning(
                f"The package wrongfully added the test folder as a module ({module}),"
                f" as a result that might result in conda clobber warnings."
            )
            continue
        result.append(module)
    if not result:
        return [impt.replace("/", ".") for impt in sorted(meta_pkg)[:2]]
    return result


def get_entry_points_from_sdist(sdist_metadata: dict) -> List:
    """Extract entry points from sdist metadata

    :param sdist_metadata: sdist metadata
    :return: list with all entry points
    """
    all_entry_points = sdist_metadata.get("entry_points", {})
    if not all_entry_points:
        return []
    if isinstance(all_entry_points, str):
        all_lines = []
        for line in all_entry_points.splitlines():
            if "=" not in line:
                all_lines.append(line)
                continue
            all_parts = [line_col.strip() for line_col in line.split("=")]
            if not all_parts[-1].startswith(("'", '"')):
                all_parts[-1] = f"'{all_parts[-1]}'"
            all_lines.append("=".join(all_parts))
        try:
            all_entry_points = tomli.loads("\n".join(all_lines))
        except tomli.TOMLDecoderError:
            return []

    if all_entry_points.get("console_scripts") or all_entry_points.get("gui_scripts"):
        console_scripts = all_entry_points.get("console_scripts", [])
        if isinstance(console_scripts, dict):
            console_scripts = [
                f"{k} = {v}" for k, v in all_entry_points["console_scripts"].items()
            ]

        gui_scripts = all_entry_points.get("gui_scripts", [])
        if isinstance(gui_scripts, dict):
            gui_scripts = [
                f"{k} = {v}" for k, v in all_entry_points["gui_scripts"].items()
            ]

        entry_points_result = []
        if console_scripts:
            if isinstance(console_scripts, str):
                console_scripts = [console_scripts]
            entry_points_result += console_scripts
        if gui_scripts:
            if isinstance(gui_scripts, str):
                gui_scripts = [gui_scripts]
            entry_points_result += gui_scripts
        return_entry_point = []
        for entry_point in entry_points_result:
            if isinstance(entry_point, str):
                entry_point = entry_point.split("\n")
            return_entry_point.extend(entry_point)
        return [ep for ep in return_entry_point if ep.strip()]
    return []


def download_sdist_pkg(sdist_url: str, dest: str, name: Optional[str] = None):
    """Download the sdist package

    :param sdist_url: sdist url
    :param dest: Folder were the method will download the sdist
    """
    print_msg(
        f"{Fore.GREEN}Starting the download of the sdist package"
        f" {Fore.BLUE}{Style.BRIGHT}{name}"
    )
    log.debug(f"Downloading {name} sdist - {sdist_url}")
    response = requests.get(sdist_url, allow_redirects=True, stream=True, timeout=5)
    response.raise_for_status()
    total_size = int(response.headers.get("Content-length", 0))
    with manage_progressbar(max_value=total_size, prefix=f"{name} ") as bar, open(
        dest, "wb"
    ) as pkg_file:
        progress_val = 0
        chunk_size = 512
        for chunk_data in response.iter_content(chunk_size=chunk_size):
            if chunk_data:
                pkg_file.write(chunk_data)
                progress_val += chunk_size
                bar.update(min(progress_val, total_size))


def get_sdist_metadata(
    sdist_url: str, config: Configuration, with_source: bool = False
) -> dict:
    """Method responsible to return the sdist metadata which is basically
    the metadata present in setup.py and setup.cfg or PKG-INFO
    :param sdist_url: URL to the sdist package
    :param config: package configuration
    :param with_source: a boolean value to indicate Github packages
    :return: sdist metadata
    """
    temp_folder = mkdtemp(prefix=f"grayskull-{config.name}-")
    if config.from_local_sdist:
        path_pkg = Path(config.local_sdist).resolve()
    else:
        pkg_name = pkg_name_from_sdist_url(sdist_url)
        path_pkg = os.path.join(temp_folder, pkg_name)

        download_sdist_pkg(sdist_url=sdist_url, dest=path_pkg, name=config.name)
        if config.download:
            config.files_to_copy.append(path_pkg)
    log.debug(f"Unpacking {path_pkg} to {temp_folder}")
    shutil.unpack_archive(path_pkg, temp_folder)
    print_msg("Recovering information from setup.py")
    with injection_distutils(temp_folder) as metadata:
        metadata["sdist_path"] = temp_folder

    # At this point the tarball was successfully extracted
    # so we can assume the sha256 can be computed reliably
    if with_source:
        metadata["source"] = {"url": sdist_url, "sha256": sha256_checksum(path_pkg)}
    if config.from_local_sdist:
        metadata["source"] = {
            "url": f"file://{path_pkg}",
            "sha256": sha256_checksum(path_pkg),
        }

    # Get some keys from PKG-INFO
    path_pkg_info = list(Path(temp_folder).rglob("PKG-INFO"))
    if path_pkg_info:
        dist = UnpackedSDist(path_pkg_info[0].parent)
        for key in ("name", "version", "summary", "author"):
            metadata[key] = getattr(dist, key, None)
    return metadata


def ensure_pep440_in_req_list(list_req: List[str]) -> List[str]:
    return [ensure_pep440(pkg) for pkg in list_req]


def ensure_pep440(pkg: str) -> str:
    if not pkg:
        return pkg
    if pkg.strip().startswith("<{") or pkg.strip().startswith("{{"):
        return pkg
    split_pkg = pkg.strip().split(" ")
    if len(split_pkg) <= 1:
        return pkg
    constrain_pkg = "".join(split_pkg[1:])
    list_constrains = constrain_pkg.split(",")
    full_constrain = []
    for constrain in list_constrains:
        if "~=" in constrain:
            version = constrain.strip().replace("~=", "").strip()
            version_upper = next_incompatible_version(version)
            full_constrain.append(f">={version},<{version_upper}")
        else:
            full_constrain.append(constrain.strip())
    all_constrains = ",".join(full_constrain)
    return f"{split_pkg[0]} {all_constrains}"


def next_incompatible_version(version: str) -> str:
    """Return the next incompatible version for the given version."""
    comments_removed = version.split("#")[0]
    version_ = Version(comments_removed)
    epoch = version_.epoch
    previous_release = version_.release
    if len(previous_release) < 2:
        raise ValueError(
            f"~= operator requires at least two version numbers, but given only "
            f"'{version}'"
        )
    release = previous_release[:-2] + (previous_release[-2] + 1,)
    release_str = ".".join(str(r) for r in release)
    return canonicalize_version(f"{epoch}!{release_str}dev")



import pytest
from souschef.jinja_expression import get_global_jinja_var

from grayskull.__main__ import create_python_recipe
from grayskull.strategy.pypi import adjust_source_url_to_include_placeholders


@pytest.mark.github
def test_289_github_url_version_placeholder():
    recipe, _ = create_python_recipe(
        "https://github.com/spdx/spdx-license-matcher", version="2.1"
    )
    assert (
        recipe["source"]["url"]
        == "https://github.com/spdx/spdx-license-matcher/archive/v{{ version }}.tar.gz"
    )
    assert get_global_jinja_var(recipe, "version") == "2.1"


def test_adjust_source_url_to_include_placeholders():
    assert (
        adjust_source_url_to_include_placeholders(
            "https://github.com/spdx/spdx-license-matcher/archive/v2.1.tar.gz", "2.1"
        )
        == "https://github.com/spdx/spdx-license-matcher/archive/v{{ version }}.tar.gz"
    )



from dataclasses import dataclass, field
from typing import Dict, List, Optional, Tuple

from grayskull.cli.parser import parse_pkg_name_version
from grayskull.utils import PyVer


@dataclass
class Configuration:
    name: str
    version: str = ""
    files_to_copy: List = field(default_factory=list)
    supported_py: List[PyVer] = field(
        default_factory=lambda: [
            PyVer(2, 7),
            PyVer(3, 6),
            PyVer(3, 7),
            PyVer(3, 8),
            PyVer(3, 9),
            PyVer(3, 10),
        ]
    )
    py_cf_supported: List[PyVer] = field(
        default_factory=lambda: [
            PyVer(3, 6),
            PyVer(3, 7),
            PyVer(3, 8),
            PyVer(3, 9),
            PyVer(3, 10),
        ]
    )
    is_strict_cf: bool = False
    pkg_need_c_compiler: Tuple = field(default_factory=lambda: ("cython",))
    pkg_need_cxx_compiler: Tuple = field(default_factory=lambda: ("pybind11",))
    url_pypi_metadata: str = "https://pypi.org/pypi/{pkg_name}/json"
    download: bool = False
    is_arch: bool = False
    repo_github: Optional[str] = None
    from_local_sdist: bool = False
    local_sdist: Optional[str] = None
    missing_deps: set = field(default_factory=set)
    extras_require_test: Optional[str] = None
    github_release_tag: Optional[str] = None
    extras_require_include: Tuple[str] = tuple()
    extras_require_exclude: Tuple[str] = tuple()
    extras_require_all: bool = False
    extras_require_split: bool = False

    def get_oldest_py3_version(self, list_py_ver: List[PyVer]) -> PyVer:
        list_py_ver = sorted(list_py_ver)
        min_python_version = (
            self.py_cf_supported[0] if self.is_strict_cf else PyVer(3, 0)
        )
        for py_ver in list_py_ver:
            if py_ver >= min_python_version:
                return py_ver
        return min_python_version

    def get_py_version_available(
        self, req_python: List[Tuple[str, str, str]]
    ) -> Dict[PyVer, bool]:
        """Get the python version available given the requires python received

        :param req_python: Requires python
        :return: Dict of Python versions if it is enabled or disabled
        """
        sup_python_ver = set(
            self.py_cf_supported if self.is_strict_cf else self.supported_py
        )
        sup_python_ver.update(
            {
                PyVer(int(major), int(minor or 0))
                for _, major, minor in req_python
                if major
            }
        )
        sup_python_ver = sorted(list(sup_python_ver))
        if self.is_strict_cf:
            py_ver_enabled = {
                py_ver: py_ver in self.py_cf_supported for py_ver in sup_python_ver
            }
        else:
            py_ver_enabled = {py_ver: True for py_ver in sup_python_ver}
        for op, major, minor in req_python:
            if op == "=":
                op = "=="
            elif op == "~=":
                op = ">="
            if not minor:
                minor = 0
            for sup_py, is_enabled in py_ver_enabled.items():
                if is_enabled is False:
                    continue
                py_ver_enabled[sup_py] = eval(
                    f"sup_py {op} PyVer(int({major}), int({minor}))"
                )
        return py_ver_enabled

    def __post_init__(self):
        if self.from_local_sdist:
            self.local_sdist = self.name
        if self.url_pypi_metadata != "https://pypi.org/pypi/{pkg_name}/json":
            preffix = "" if self.url_pypi_metadata.endswith("/") else "/"
            self.url_pypi_metadata += f"{preffix}{{pkg_name}}/json"
        pkg_repo, pkg_name, pkg_version = parse_pkg_name_version(self.name)
        if pkg_repo:
            prefix = "" if pkg_repo.endswith("/") else "/"
            self.repo_github = f"{pkg_repo}{prefix}{pkg_name}"

        self.name = pkg_name

        if pkg_version:
            self.version = pkg_version



from progressbar import ProgressBar

from grayskull.cli import CLIConfig
from grayskull.cli.stdout import manage_progressbar, print_msg


def test_print_stdout(capsys):
    CLIConfig().stdout = True
    print_msg("TEST-OUTPUT")
    captured_out = capsys.readouterr()
    assert captured_out.out == "TEST-OUTPUT\n"


def test_disabled_print(capsys):
    CLIConfig().stdout = False
    print_msg("TEST-OUTPUT")
    captured_out = capsys.readouterr()
    assert captured_out.out == ""


def test_progressbar_enable():
    CLIConfig().stdout = True
    with manage_progressbar(max_value=100, prefix="prefix-") as bar:
        assert isinstance(bar, ProgressBar)


def test_progressbar_disable():
    CLIConfig().stdout = False
    with manage_progressbar(max_value=100, prefix="prefix-") as bar:
        assert not isinstance(bar, ProgressBar)



import os
from typing import List
from unittest.mock import patch

import pytest
import requests
from pytest import fixture

from grayskull.license.discovery import (
    _get_all_license_choice,
    _get_all_names_from_api,
    _get_api_github_url,
    _get_license,
    get_all_licenses_from_spdx,
    get_license_type,
    get_opensource_license_data,
    get_other_names_from_opensource,
    get_short_license_id,
    match_license,
    search_license_api_github,
    search_license_folder,
    search_license_repo,
)


@fixture
def license_pytest_path(data_dir) -> str:
    return os.path.join(data_dir, "licenses", "pytest.txt")


@fixture
def spdx_org_license_mit() -> List:
    return [
        {
            "reference": "./MIT.html",
            "isDeprecatedLicenseId": False,
            "isFsfLibre": True,
            "detailsUrl": "http://spdx.org/licenses/MIT.json",
            "referenceNumber": "256",
            "name": "MIT License",
            "licenseId": "MIT",
            "seeAlso": ["https://opensource.org/licenses/MIT"],
            "isOsiApproved": True,
        }
    ]


def test_match_license():
    assert match_license("MIT License")["licenseId"] == "MIT"
    assert match_license("Expat")["licenseId"] == "MIT"


def test_get_all_licenses_from_spdx():
    assert len(get_all_licenses_from_spdx()) > 300


def test_get_opensource_license_data():
    assert len(get_opensource_license_data()) >= 50


def test_short_license_id():
    assert get_short_license_id("MIT License") == "MIT"
    assert get_short_license_id("Expat") == "MIT"
    assert get_short_license_id("GPL 2.0") == "GPL-2.0-or-later"
    assert get_short_license_id("2-Clause BSD License") == "BSD-2-Clause"
    assert get_short_license_id("3-Clause BSD License") == "BSD-3-Clause"


def test_get_other_names_from_opensource():
    assert sorted(get_other_names_from_opensource("MIT")) == sorted(["MIT", "Expat"])


def test_get_license(spdx_org_license_mit):
    assert _get_license("MIT", spdx_org_license_mit) == spdx_org_license_mit[0]


def test_get_all_names_from_api(spdx_org_license_mit):
    assert sorted(_get_all_names_from_api(spdx_org_license_mit[0])) == sorted(
        ["Expat", "MIT", "MIT License"]
    )


def test_get_all_license_choice(spdx_org_license_mit):
    assert sorted(_get_all_license_choice(spdx_org_license_mit)) == sorted(
        ["Expat", "MIT", "MIT License"]
    )


@fixture
def license_pytest_5_3_1(license_pytest_path) -> str:
    with open(license_pytest_path) as f:
        return f.read()


@pytest.mark.github
def test_search_license_api_github(license_pytest_5_3_1: str):
    license_api = search_license_api_github(
        "https://github.com/pytest-dev/pytest", "5.3.1"
    )
    assert license_api.name == "MIT"
    assert license_api.path.endswith("LICENSE")

    with open(license_api.path, "r") as f:
        assert f.read() == license_pytest_5_3_1


@pytest.mark.github
def test_get_api_github_url():
    assert (
        _get_api_github_url("https://github.com/pytest-dev/pytest", "5.3.1")
        == "https://api.github.com/repos/pytest-dev/pytest/license?ref=5.3.1"
    )
    assert (
        _get_api_github_url("https://github.com/pytest-dev/pytest")
        == "https://api.github.com/repos/pytest-dev/pytest/license"
    )


def test_search_license_folder(pkg_pytest):
    license_folder = search_license_folder(os.path.dirname(pkg_pytest))[0]
    assert license_folder.path == os.path.join(
        os.path.dirname(pkg_pytest), "pytest-5.3.5", "LICENSE"
    )
    assert license_folder.name == "MIT"


@pytest.mark.xfail(
    reason="This test may fail because github has limitation regarding the"
    " number of requisitions we can do to their api."
)
def test_search_license_repository(pkg_pytest):
    license_repo = search_license_repo("https://github.com/pytest-dev/pytest", "5.3.5")
    assert license_repo.path.endswith("LICENSE")
    assert license_repo.name == "MIT"


def test_predict_license_type(license_pytest_path):
    assert get_license_type(license_pytest_path) == "MIT"


def test_fallback_cache_licence():
    with patch("requests.get", side_effect=requests.exceptions.RequestException):
        assert get_opensource_license_data()



import os

from setuptools import find_packages, setup

if os.path.exists("README.md"):
    with open("README.md", "r") as f:
        readme = f.read()
else:
    readme = ""


setup(
    name="grayskull",
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "grayskull = grayskull.__main__:main",
            "greyskull = grayskull.__main__:main",
            "conda-grayskull = grayskull.__main__:main",
            "conda-greyskull = grayskull.__main__:main",
        ]
    },
    use_scm_version={"write_to": "grayskull/_version.py"},
    setup_requires=["setuptools-scm", "setuptools>=30.3.0"],
    package_data={
        "": [
            "LICENSE",
            "license/data/*",
            "license/data/*.*",
            "licence/licence_cache.json",
        ]
    },
    include_package_data=True,
    python_requires=">=3.8",
    install_requires=[
        "requests",
        "ruamel.yaml >=0.16.10",
        "ruamel.yaml.jinja2",
        "stdlib-list",
        "pip",
        "setuptools >=30.3.0",
        "rapidfuzz >=1.7.1",
        "progressbar2 >=3.53.0",
        "colorama",
        "conda-souschef >=2.2.1",
        "pkginfo",
        "packaging >=21.3",
        "tomli",
        "tomli-w",
    ],
    extras_require={
        "testing": [
            "pytest",
            "mock",
            "pytest-cov",
            "pytest-console-scripts",
            "pytest-mock",
            "setuptools-scm",
        ]
    },
    url="https://github.com/conda-incubator/grayskull",
    license="MIT",
    author="Marcelo Duarte Trevisani",
    author_email="marceloduartetrevisani@gmail.com",
    description="Project to generate recipes for conda packages.",
    long_description_content_type="text/markdown",
    long_description=readme,
    project_urls={"Source": "https://github.com/conda-incubator/grayskull"},
)
